"""
Command classes for Distutils/setup.py to assist with Python package building/usage/management.

For basic :class:`.BumpCommand` usage, see either the class documentation itself, or the :py:mod:`.bump` module
documentation.

**Registering distutils commands in setup.py**

Below is a very basic setup.py showing how to load the commands :class:`.BumpCommand` and :class:`.ExtrasCommand`
into setup.py, so they can be used via ``./setup.py bump`` and ``./setup.py extras`` respectively.

.. code-block:: python
        
        from setuptools import setup, find_packages
        from privex.helpers import settings, BumpCommand, ExtrasCommand
        
        # For BumpCommand, specify the file which contains "VERSION = '1.2.3'"
        settings.VERSION_FILE = '/home/john/mypackage/mypackage/__init__.py'
        
        # For ExtrasCommand / extras_require, you may wish to change EXTRAS_FOLDER if you want to store your
        # extras requirements.txt files in a different folder structure.
        # By default, EXTRAS_FOLDER is 'extras' (a folder, relative to the current working directory)
        settings.EXTRAS_FOLDER = 'requirements/extras'
        
        # Load the extra requirements files requirements/extras/cache.txt and requirements/extras/tests.txt
        extensions = ['cache', 'tests']
        
        setup(
            extras_require=extras_require(extensions),
            cmdclass={
                'bump': BumpCommand,
                'extras': ExtrasCommand,
            },
        );

"""
import os
import subprocess
import sys
from distutils.cmd import Command
from io import BufferedWriter
from tempfile import NamedTemporaryFile
from typing import Optional

from privex.helpers import settings
from privex.helpers.setuppy import extras, EXTRAS_FOLDER


class BumpCommand(Command):
    """
    Distutils/setup.py command class for bumping the package version using :func:`.bump_version`
    
    **Usage in setup.py:**
    
    .. code-block:: python
            
            from setuptools import setup, find_packages
            from privex.helpers import settings, BumpCommand
            
            # The file which contains "VERSION = '1.2.3'"
            settings.VERSION_FILE = '/home/john/mypackage/mypackage/__init__.py'
            
            setup(
                cmdclass={
                    'bump': BumpCommand
                },
            );
    
    **Usage on the command line:**
    
    .. code-block:: bash
    
        ./setup.py bump --patch
        # Bumping 'patch' version part
        # Updating version stored in file @ /tmp/helpers/privex/helpers/__init__.py
        # Package version has been bumped from 2.0.0 to 2.0.1 and written to the file
        # /tmp/helpers/privex/helpers/__init__.py
        
        ./setup.py bump --minor
        # ... version has been bumped from 2.0.0 to 2.1.0 and written to the file ...
        
        ./setup.py bump --pre
        # ... version has been bumped from 2.1.0 to 2.1.0-rc.1 ...
        
        ./setup.py bump --build
        # ... version has been bumped from 2.2.0 to 2.2.0+build.1 ...
    
    The options ``--pre`` and ``--build`` also allow an extra option ``--token=xxx`` which allows you to
    change the **token** from the respective ``-rc.x`` and ``+build.x`` to your own label.
    
    If a token is found in the version string, then it will automatically use the one in the version string
    until you either clean off the pre/build suffixes with a normal version bump e.g. ``--patch``
    or you manually remove the suffixes from the version number in the file.
    
    .. code-block:: bash
        
        # NOTE: to change the suffix token, the version must be plain x.y.z without -xxx.y appended.
        ./setup.py bump --pre --token=alpha
        # ... version has been bumped from 2.1.0 to 2.1.0-alpha.1 ...

        ./setup.py bump --pre
        # ... version has been bumped from 2.1.0-alpha.1 to 2.1.0-alpha.2 ...
       
    
    """
    description = 'Bumps the package version using privex.helpers.setuppy.bump_version'
    user_options = [
        ('dry', None, 'If specified, will output the updated file contents, but will not write them'),
        ('patch', None, 'Bump patch version (0.0.x)'),
        ('minor', None, 'Bump minor version (0.x.0'),
        ('major', None, 'Bump major version (x.0.0)'),
        ('build', None, 'Bump build version (0.0.0+build.x)'),
        ('pre', None, 'Bump prerelease version (0.0.0-rc.x)'),
        ('token=', None, 'For build/pre, use this argument value instead of "build" / "rc"'),
    ]
    """Option arguments which can be specified by the user on the command line for this command"""
    
    patch: Optional[int]
    """Bump patch version. If ``--patch`` is passed, this should change from ``None`` to ``1``"""
    
    minor: Optional[int]
    """Bump minor version. If ``--minor`` is passed, this should change from ``None`` to ``1``"""
    
    major: Optional[int]
    """Bump major version. If ``--major`` is passed, this should change from ``None`` to ``1``"""
    
    build: Optional[int]
    """Bump build version. If ``--build`` is passed, this should change from ``None`` to ``1``"""
    
    pre: Optional[int]
    """Bump prerelease version. If ``--pre`` is passed, this should change from ``None`` to ``1``"""
    
    token: Optional[str]
    """
    If bumping :py:attr:`.pre` or :py:attr:`.build` then the user may specify ``--token`` to configure the
    version token e.g. ``rc`` or ``alpha``
    """
    
    dry: Optional[int]
    """
    Enable dry run mode if not ``None``. Dry mode means the contents of the file that would be modified would be
    returned to the console so you can see how it would modify the file, without actually writing to it.
    
    If ``--dry`` is specified, this should change from from ``None`` to ``1``
    """
    
    version_part: Optional[str]
    """Stores the chosen version part to bump, e.g. ``'pre'`` or ``'major'``"""
    
    def initialize_options(self):
        self.patch = None
        self.minor = None
        self.major = None
        self.build = None
        self.pre = None
        self.token = None
        self.version_part = None
        self.dry = None
    
    def finalize_options(self):
        exclusive = ['patch', 'minor', 'major', 'build', 'pre']
        ex_set = False
        for ex in exclusive:
            
            if getattr(self, ex) is not None:
                assert not ex_set, f'Error: Version portion arguments are mutually exclusive.' \
                                   f'Only pass ONE of these options: {", ".join(exclusive)}'
                ex_set = True
                self.version_part = ex
        assert ex_set, f'Please pass at least ONE of these options: --{", --".join(exclusive)}'
    
    def run(self):
        from privex.helpers.setuppy.bump import bump_version
        ver_file = settings.VERSION_FILE
        ver_args = dict(part=self.version_part)
        if self.token is not None:
            self.announce(f'Using version token "{self.token}"', level=4)
            ver_args['token'] = self.token
        if self.dry is not None:
            ver_args['dry'] = True
            self.announce("##############################################################", level=4)
            self.announce("# !!! DRY RUN !!!", level = 4)
            self.announce(f"# Modified contents of {ver_file} (not written back to file)", level=4)
            self.announce("##############################################################", level=4)
            contents = bump_version(**ver_args)
            self.announce(contents, level=4)
            self.announce("##############################################################", level=4)
            self.announce("# !!! END DRY RUN !!!", level=4)
            self.announce(f"# Above is modified contents of {ver_file} (not written back to file)", level=4)
            self.announce("##############################################################", level=4)
            return

        self.announce(f"Bumping '{self.version_part}' version part", level=4)
        self.announce(f"Updating version stored in file @ {ver_file}", level=4)
        new_version, old_version = bump_version(**ver_args)
        self.announce(f"Package version has been bumped from {old_version} to {new_version} and "
                      f"written to the file {ver_file}", level=4)


class ExtrasCommand(Command):
    """
    Distutils/setup.py command for managing package extras, including displaying/saving requirements, installing them,
    and more.
    
    **Usage in setup.py:**
    
    .. code-block:: python
            
            from setuptools import setup, find_packages
            from privex.helpers import ExtrasCommand
            
            
            setup(
                cmdclass={
                    'extras': ExtrasCommand
                },
            );
    
    **Command usage examples**
    
    Print all requirements listed for each extra in ``extras_require``
    
    .. code-block:: bash
    
        ./setup.py extras
        #### Example output: ####
        # # Extra [cache] - file: extras/cache.txt
        # redis>=3.3
        # # Extra [crypto] - file: extras/crypto.txt
        # cryptography>=2.8
        # Extra [django] - file: extras/django.txt
        # Django
        ...
    
    
    Print only the requirements for a specific extra:
    
    .. code-block:: bash
        
        ./setup.py extras --extra=cache
        #### Example output: ####
        # # Extra [cache] - file: extras/cache.txt
        # redis>=3.3
    
    
    Install requirements for a specific extra using pip
    
    
    .. code-block:: bash
    
        ./setup.py extras --install --extra=cache
        # running extras
        # Installing the following requirements: redis>=3.3
        # Requirement already up-to-date: redis>=3.3 in ./venv/lib/python3.7/site-packages (3.3.11)
        # Done.
    
    Install **ALL** extras requirements using pip
    
    .. code-block:: bash
        
        ./setup.py extras --install
    
    List extras entered in ``extras_require``
    
    .. code-block:: bash
    
        ./setup.py extras --list
        # # Extras available:
        # full
        # cache
        # crypto
        ...
    
    """
    user_options = [
        ('save=', 'o', 'Output extras in requirements.txt format to this file'),
        ('extra=', 'e', 'List requirements for this specific extra name, e.g. --extra=cache'),
        ('list', 'l', 'Only list extra names, not the requirements they include'),
        ('install', 'i', 'Install all extras requirements, or if --extra/-e is specified, only that extra.')
    ]
    description = 'Manage distutils extras. Prints/saves requirements, prints extras, and can ' \
                  'install extras via pip'
    
    def initialize_options(self):
        self.save = None
        self.extra = None
        self.list = None
        self.install = None
        return
    
    def finalize_options(self):
        return
    
    def _get_extra(self, name, comments=True) -> list:
        """
        Return a list of requirements for the extra ``name``
        
        :param str name: Name of the extra, e.g. ``'cache'``
        :param bool comments: (Default: ``True``) If True, add a comment to the start containing the extra name
                              and path to it's requirements file
        :return list reqs: The extras requirements, as a list
        """
        ex_file = name + ".txt"
        extra_list = extras(ex_file)
        if comments:
            extra_list.insert(0, f'# Extra [{name}] - file: {os.path.join(EXTRAS_FOLDER, ex_file)}')
        return extra_list

    def _get_all_extras(self, comments=True) -> list:
        """
        Return a combined list of requirements for all package extras.
        
        :param bool comments: (Default: ``True``) If True, add a comment to the start of each extras requirements
                              containing the extra name and path to it's requirements file
        :return list reqs: The extras requirements, as a list
        """
        from distutils.dist import Distribution
        dis: Distribution = self.distribution
        ext = dis.metadata.provides_extras
        extra_list = []
        for ex in ext:
            extra_list = extra_list + self._get_extra(ex, comments=comments)
        return extra_list
    
    def install_extras(self):
        """
        Installs extras requirements using ``python -m pip install -U -r requirements.txt``
        
        The python executable is determined using ``sys.executable``, and the requirements.txt file is generated
        into a temporary file which is deleted after the install is finished.
        
        If the argument ``--extra`` / ``-e``  ( :py:attr:`.extra` ) is passed, then this will install
        only the requirements for that extra.
        
        **Example** (assuming this class is registered as ``extras``)
        
        .. code-block:: bash
            
            # Install ALL extras requirements
            ./setup.py extras --install
            
            # Install only the requirements for the extra 'cache'
            ./setup.py extras --install --extra=cache
            # Shorthand version (YMMV based on setuptools/distutils version)
            ./setup.py extras -i -e cache
        
            
        """
        with NamedTemporaryFile() as req_file:
            extras_list = self._get_extra(self.extra, False) if self.extra else self._get_all_extras(False)
            print('Installing the following requirements:', ', '.join(extras_list))
            self.save_list(extra_list=extras_list, out_file=req_file.name)
            subprocess.check_call([sys.executable, '-m', 'pip', 'install', '-U', '-r', req_file.name])

            print('Done.')

    def run(self):
        if self.install:
            return self.install_extras()
        
        from distutils.dist import Distribution
        dis: Distribution = self.distribution
        ext = dis.metadata.provides_extras
        if self.list:
            print("# Extras available:\n")
            for ex in ext:
                print(ex)
            return
        if self.extra:
            extra_list = self._get_extra(self.extra)
            if self.save:
                self.save_list(extra_list)
                self.announce(f"Wrote extra '{self.extra}' requirements to file {self.save}", level=4)
                return
            for ex in extra_list:
                print(ex)
            return
        
        extra_list = self._get_all_extras()
        if self.save:
            self.save_list(extra_list)
            self.announce(f"Wrote all extras requirements to file {self.save}", level=4)
            return
        
        # ex_file = ex + ".txt"
        # print(f'# Extra [{ex}] - file: {os.path.join(EXTRAS_FOLDER, ex_file)}')
        for r in extra_list:
            print(r)

    def save_list(self, extra_list, out_file=None):
        out_file = self.save if not out_file else out_file
        if isinstance(out_file, BufferedWriter):
            for ex in extra_list:
                out_file.write(ex + "\n")
            return
        
        with open(out_file, 'w') as fp:
            for ex in extra_list:
                fp.write(ex + "\n")



import logging
from datetime import datetime, timedelta
from typing import Any, Optional, Union

from privex.helpers.exceptions import CacheNotFound
from privex.helpers.settings import DEFAULT_CACHE_TIMEOUT
from privex.helpers.cache.CacheAdapter import CacheAdapter

log = logging.getLogger(__name__)


class MemoryCache(CacheAdapter):
    """
    A very basic cache adapter which implements :class:`.CacheAdapter` - stores the cache in memory using
    the static attribute :py:attr:`.__CACHE`
    
    As the cache is simply stored in memory, any python object can be cached without needing any form of serialization.
    
    Fully supports cache expiration.
    
    **Basic Usage**::
        
        >>> from time import sleep
        >>> c = MemoryCache()
        >>> c.set('test:example', 'hello world', timeout=60)
        >>> c.get('test:example')
        'hello world'
        >>> sleep(60)
        >>> c.get('test:example', 'NOT FOUND')
        'NOT FOUND'
    
    """
    __CACHE = {}
    
    def get(self, key: str, default: Any = None, fail: bool = False) -> Any:
        key = str(key)
        c = self.__CACHE
        if key in c:
            log.debug('Cache key "%s" found in __CACHE. Checking expiry...', key)
            vc = c[key]
            if str(vc['timeout']) != 'never' and vc['timeout'] < datetime.utcnow():
                log.debug('Cache key "%s" has expired. Removing from cache.')
                del c[key]
                if fail:
                    raise CacheNotFound(f'Cache key "{key}" was expired.')
                return default
            log.debug('Cache key "%s" is valid and not expired. Returning value "%s"', key, vc)
            return vc['value']
        if fail:
            raise CacheNotFound(f'Cache key "{key}" was not found.')
        log.debug('Cache key "%s" was not found in __CACHE. Returning default value.', key)
        return default

    def set(self, key: str, value: Any, timeout: Optional[int] = DEFAULT_CACHE_TIMEOUT):
        key, timeout = str(key), int(timeout)
        c = self.__CACHE
        log.debug('Setting cache key "%s" to value "%s" with timeout %s', key, value, timeout)
        c[key] = dict(value=value, timeout=datetime.utcnow() + timedelta(seconds=timeout))
        return c[key]

    def get_or_set(self, key: str, value: Union[Any, callable], timeout: int = DEFAULT_CACHE_TIMEOUT) -> Any:
        key, timeout = str(key), int(timeout)
        try:
            k = self.get(key, fail=True)
        except CacheNotFound:
            k = value(key) if callable(value) else value
            self.set(key=key, value=k, timeout=timeout)
        return k
    
    def remove(self, *key: str) -> bool:
        removed = 0
        for k in key:
            k = str(k)
            if k in self.__CACHE:
                del self.__CACHE[k]
                removed += 1
        return removed == len(key)

    def update_timeout(self, key: str, timeout: int = DEFAULT_CACHE_TIMEOUT) -> Any:
        key, timeout = str(key), int(timeout)
        v = self.get(key=key, fail=True)
        self.__CACHE[key]['timeout'] = datetime.utcnow() + timedelta(seconds=timeout)
        return v



"""
This module file contains Django-specific helper functions, to help save time when 
developing with the Django framework.


 * handle_error              - Redirects normal web page requests with a session error, outputs JSON with a 
   status code for API queries.

 * is_database_synchronized  - Check if all migrations have been ran before running code.

 * model_to_dict             - Extract an individual Django model instance into a dict (with display names)
 
 * to_json                   - Convert a model Queryset into a plain string JSON array with display names


**Copyright**::

        +===================================================+
        |                 © 2019 Privex Inc.                |
        |               https://www.privex.io               |
        +===================================================+
        |                                                   |
        |        Originally Developed by Privex Inc.        |
        |        License: X11 / MIT                         |
        |                                                   |
        |        Core Developer(s):                         |
        |                                                   |
        |          (+)  Chris (@someguy123) [Privex]        |
        |          (+)  Kale (@kryogenic) [Privex]          |
        |                                                   |
        +===================================================+

    Copyright 2019     Privex Inc.   ( https://www.privex.io )



"""
import json
from django.contrib import messages
from django.contrib.messages import add_message
from django.http import JsonResponse, HttpRequest
from django.http.response import HttpResponseRedirectBase
from django.db.migrations.executor import MigrationExecutor
from django.db import connections

__all__ = ['handle_error', 'is_database_synchronized', 'model_to_dict', 'to_json']


def handle_error(request: HttpRequest, err: str, rdr: HttpResponseRedirectBase, status=400):
    """
    Output an error as either a Django session message + redirect, or a JSON response
    based on whether the request was for the API readable version (?format=json) or not.

    Usage:

    >>> from django.shortcuts import redirect
    >>> def my_view(request):
    ...     return handle_error(request, "Invalid password", redirect('/login'), 403)

    :param HttpRequest request:  The Django request object from your view
    :param str err:              An error message as a string to display to the user / api call
    :param HttpResponseRedirectBase rdr:  A redirect() for normal browsers to follow after adding the session error.
    :param int status:           The HTTP status code to return if the request is an API call (default: 400 bad request)
    """
    if request.GET.get('format', '') == 'json':
        return JsonResponse(dict(error=True, message=err), status=status)
    else:
        add_message(request, messages.ERROR, err)
        return rdr


def is_database_synchronized(database: str) -> bool:
    """
    Check if all migrations have been ran. Useful for preventing auto-running code accessing models before the
    tables even exist, thus preventing you from migrating...

    >>> from django.db import DEFAULT_DB_ALIAS
    >>> if not is_database_synchronized(DEFAULT_DB_ALIAS):
    >>>     log.warning('Cannot run reload_handlers because there are unapplied migrations!')
    >>>     return

    :param str database: Which Django database config is being used? Generally just pass django.db.DEFAULT_DB_ALIAS
    :return bool: True if all migrations have been ran, False if not.
    """

    connection = connections[database]
    connection.prepare_database()
    executor = MigrationExecutor(connection)
    targets = executor.loader.graph.leaf_nodes()
    return False if executor.migration_plan(targets) else True


def model_to_dict(model) -> dict:
    """1 dimensional json-ifyer for any Model"""
    from django.forms import model_to_dict as mtd
    # gets a field on the model, using the display name if available
    def get_display(model, field, default):
        method_name = 'get_{}_display'.format(field)
        return getattr(model, method_name)() if method_name in dir(model) else default

    return {k: get_display(model, k, v) for k, v in mtd(model).items()}


def to_json(query_set) -> str:
    """Iterate a Django query set and dump to json str"""
    return json.dumps([model_to_dict(e) for e in query_set], default=str)



"""
Cryptography related helper classes/functions

**Dependencies**

Requires the ``cryptography`` Python package:

.. code-block:: bash

    # Either install privex-helpers with the `crypto` extra
    pipenv install 'privex-helpers[crypto]'  # Using pipenv if you have it
    pip3 install 'privex-helpers[crypto]'    # Using standard pip3
    
    # Or manually install the `cryptography` library.
    pipenv install cryptography   # Using pipenv if you have it
    pip3 install cryptography     # Using standard pip3

**Summary**

Some of the most useful parts of this module include :py:mod:`.EncryptHelper` and :py:mod:`.KeyManager` - these
two components cover both symmetric (shared key) encryption/decryption, as well as asymmetric (public/private key)
keypair generation, signing, verification, as well as encryption/decryption with RSA.


 * :py:mod:`.EncryptHelper` - Painless symmetric encryption / decryption with AES-128
 * :py:mod:`.KeyManager` - Painless generation of asymmetric keys, with signing/verification and en/decryption support


**EncryptHelper - Painless symmetric encryption with AES-128**

:py:mod:`.EncryptHelper` is a wrapper class designed to make it extremely easy to use asymmetric encryption and
decryption. At it's core is the :py:class:`cryptography.fernet.Fernet` encryption system, and the class is
designed to make usage of Fernet as painless as possible.

**Basic Usage of EncryptHelper**::

    >>> from privex.helpers import EncryptHelper
    >>> key = EncryptHelper.generate_key() # Generates a 32-byte symmetric key, returned as a base64 encoded string
    >>> key_out = EncryptHelper.generate_key('my_key.txt')  # Generates and saves a key to my_key.txt, then returns it.
    >>> crypt = EncryptHelper(key)         # Create an instance of EncryptHelper, en/decrypting using ``key`` by default
    # Encrypts the string 'hello world' with AES-128 CBC using the instance's key, returned as a base64 string
    >>> enc = crypt.encrypt_str('hello world')
    >>> print(enc)
    gAAAAABc7ERTpu2D_uven3l-KtU_ewUC8YWKqXEbLEKrPKrKWT138MNq-I9RRtCD8UZLdQrcdM_IhUU6r8T16lQkoJZ-I7N39g==

    >>> crypt.is_encrypted(enc)       # Check if a string/bytes is encrypted (only works with data matching the key)
    True
    >>> data = crypt.decrypt_str(enc) # Decrypt the encrypted data using the same key, outputs as a string
    >>> print(data)
    hello world

**KeyManager - Painless generation of asymmetric keys, with signing/verification and en/decryption support**

While the :doc:`Cryptography <cryptography:index>` library is a brilliant library with many features, and a good
security track record - it's asymmetric key features require a ridiculous amount of scaffolding to make them usable
in a project.

That's where :py:mod:`.KeyManager` comes in.

Watch how simple it is to generate and save two types of asymmetric keys (RSA and Ed25519), and put them
to use with signatures and encryption.

First, let's generate a keypair and save it to disk. Without any algorithm related configuration options, it will
simply generate an RSA 2048 private and public key, and save them to the current working directory. For convenience,
it will also return the private and public key, so you can make use of them immediately after generating them.

    >>> rsa_priv, rsa_pub = KeyManager.output_keypair('id_rsa', 'id_rsa.pub')

Now, let's make an Ed25519 key - an algorithm quickly becoming common for SSH keys thanks to it's extremely small
public + private keys that could practically fit in an SMS message.

    >>> ed_priv, ed_pub = KeyManager.output_keypair('id_ed25519', 'id_ed25519.pub', alg='ed25519')

To use the signature / encryption functionality, first we have to load a key into :class:`.KeyManager`. There's
two ways you can do this. The first is simply passing the key as a string/bytes into the constructor. The second is
loading the key from disk (useful if you're using an existing key you have saved as a file).

Loading keys::

    # Example 1. Let's just pass our Ed25519 private key bytes straight into KeyManager.
    >>> km_ed = KeyManager(ed_priv)
    # Example 2. For the RSA key, we'll load it from a file
    >>> km_rsa = KeyManager.load_keyfile('id_rsa')

Now let's sign a message with each key, and then verify the message::

    # Sign the message "hello world" with both our Ed25519 and RSA private key.
    >>> msg = 'hello world'
    >>> sig_ed = km_ed.sign(msg)
    >>> sig_rsa = km_rsa.sign(msg)
    # Now we can verify the signature using the public keys (automatically re-generated from the private key)
    >>> km_ed.verify(signature=sig_ed, message=msg)
    True
    >>> km_rsa.verify(signature=sig_rsa, message=msg)
    True
    # If the signature was invalid, e.g. if we pass the RSA signature to the Ed25519 KeyManager...
    >>> km_ed.verify(sig_rsa, 'hello world')
        Traceback (most recent call last):
          File "<input>", line 1, in <module>
            raise InvalidSignature
        cryptography.exceptions.InvalidSignature
    # So we can see that the signatures actually work :)

For RSA keys, we can also do encryption and decryption of small messages::

    >>> enc = km_rsa.encrypt(msg)
    # For easy storage/transmission, the encrypted data is base64 encoded. This ensures you can transmit the
    # encrypted message cleanly over email, HTTP etc. without the bytes getting garbled.
    >>> enc
    b'Sf1PC_TViZdA4lq7PwSnRTLbWX20vcCtkLyQWazE9EfM9_AIn6pNTHG...
    # Now when we run `decrypt`, we get back our original message of "hello world"
    # Note: decrypt supports both raw bytes as well as base64 encoded data and will automatically detect whether
    # or not it has to decode base64.
    >>> km_rsa.decrypt(enc)
    b'hello world'

 
**Copyright**::

        +===================================================+
        |                 © 2019 Privex Inc.                |
        |               https://www.privex.io               |
        +===================================================+
        |                                                   |
        |        Originally Developed by Privex Inc.        |
        |        License: X11 / MIT                         |
        |                                                   |
        |        Core Developer(s):                         |
        |                                                   |
        |          (+)  Chris (@someguy123) [Privex]        |
        |          (+)  Kale (@kryogenic) [Privex]          |
        |                                                   |
        +===================================================+

    Copyright 2019     Privex Inc.   ( https://www.privex.io )


"""
from privex.helpers.crypto.EncryptHelper import EncryptHelper
from privex.helpers.crypto.base import is_base64, auto_b64decode
from privex.helpers.crypto.KeyManager import KeyManager, Format



"""
Common functions and classes that don't fit into a specific category

**Copyright**::

        +===================================================+
        |                 © 2019 Privex Inc.                |
        |               https://www.privex.io               |
        +===================================================+
        |                                                   |
        |        Originally Developed by Privex Inc.        |
        |        License: X11 / MIT                         |
        |                                                   |
        |        Core Developer(s):                         |
        |                                                   |
        |          (+)  Chris (@someguy123) [Privex]        |
        |          (+)  Kale (@kryogenic) [Privex]          |
        |                                                   |
        +===================================================+

    Copyright 2019     Privex Inc.   ( https://www.privex.io )

"""
import inspect
import math
import random
import re
import string
import argparse
import logging
import sys
from collections import namedtuple, OrderedDict
from datetime import datetime
from decimal import Decimal, getcontext
from os import getenv as env
from typing import Sequence, List, Union, Tuple, Type, Dict, TypeVar, Any, Iterable, Callable, NewType, Optional

from privex.helpers.collections import DictObject, OrderedDictObject

log = logging.getLogger(__name__)

SAFE_CHARS = 'abcdefhkmnprstwxyz23456789ACDEFGHJKLMNPRSTWXYZ'
"""Characters that shouldn't be mistaken, avoiding users confusing an o with a 0 or an l with a 1 or I"""

ALPHANUM = string.ascii_uppercase + string.digits + string.ascii_lowercase
"""All characters from a-z, A-Z, and 0-9 - for random strings where there's no risk of user font confusion"""

T = TypeVar('T')
"""Plain generic type variable for use in helper functions"""
K = TypeVar('K')
"""Plain generic type variable for use in helper functions"""
V = TypeVar('V')
"""Plain generic type variable for use in helper functions"""

C = TypeVar('C', type, callable, Callable)
"""Generic type variable constrained to :class:`type` / :class:`typing.Callable` for use in helper functions"""
CL = TypeVar('CL', type, callable, Callable)
"""Generic type variable constrained to :class:`type` / :class:`typing.Callable` for use in helper functions"""


def random_str(size: int = 50, chars: Sequence = SAFE_CHARS) -> str:
    """
    Generate a random string of arbitrary length using a given character set (string / list / tuple). Uses Python's 
    SystemRandom class to provide relatively secure randomness from the OS. (On Linux, uses /dev/urandom)

    By default, uses the character set :py:attr:`.SAFE_CHARS` which contains letters a-z / A-Z and numbers 2-9
    with commonly misread characters removed (such as ``1``, ``l``, ``L``, ``0`` and ``o``). Pass 
    :py:attr:`.ALPHANUM` as `chars` if you need the full set of upper/lowercase + numbers.

    Usage:

        >>> from privex.helpers import random_str
        >>> # Default random string - 50 character alphanum without easily mistaken chars
        >>> password = random_str()
        'MrCWLYMYtT9A7bHc5ZNE4hn7PxHPmsWaT9GpfCkmZASK7ApN8r'
        >>> # Customised random string - 12 characters using only the characters `abcdef12345` 
        >>> custom = random_str(12, chars='abcdef12345')
        'aba4cc14a43d'

    Warning: As this relies on the OS's entropy features, it may not be cryptographically secure on non-Linux platforms:

    > The returned data should be unpredictable enough for cryptographic applications, though its exact quality 
    > depends on the OS implementation.

    :param int size:  Length of random string to generate (default 50 characters)
    :param str chars: Characterset to generate with ( default is :py:attr:`.SAFE_CHARS` - a-z/A-Z/0-9 with 
                      often misread chars removed)

    """
    return ''.join(random.SystemRandom().choice(chars) for _ in range(size))


def empty(v, zero: bool = False, itr: bool = False) -> bool:
    """
    Quickly check if a variable is empty or not. By default only '' and None are checked, use ``itr`` and ``zero`` to
    test for empty iterable's and zeroed variables.

    Returns ``True`` if a variable is ``None`` or ``''``, returns ``False`` if variable passes the tests

    Example usage:

        >>> x, y = [], None
        >>> if empty(y):
        ...     print('Var y is None or a blank string')
        ...
        >>> if empty(x, itr=True):
        ...     print('Var x is None, blank string, or an empty dict/list/iterable')

    :param v:    The variable to check if it's empty
    :param zero: if ``zero=True``, then return ``True`` if the variable is int ``0`` or str ``'0'``
    :param itr:  if ``itr=True``, then return ``True`` if the variable is ``[]``, ``{}``, or is an iterable and has 0 length
    :return bool is_blank: ``True`` if a variable is blank (``None``, ``''``, ``0``, ``[]`` etc.)
    :return bool is_blank: ``False`` if a variable has content (or couldn't be checked properly)
    """

    _check = [None, '']
    if zero: _check += [0, '0']
    if v in _check: return True
    if itr:
        if v == [] or v == {}: return True
        if hasattr(v, '__len__') and len(v) == 0: return True

    return False


USE_ORIG_VAR = type('UseOrigVar', (), {})
"""
A simple functionless type, used purely as a default parameter value meaning "fallback to the value from a certain
other parameter".

Primarily used in :func:`.empty_if` but can be used by any function/method, including use outside of privex-helpers.

"""


def empty_if(v: V, is_empty: K = None, not_empty: T = USE_ORIG_VAR, **kwargs) -> Union[T, K, V]:
    """
    Syntactic sugar for ``x if empty(y) else z``. If ``not_empty`` isn't specified, then the original value ``v``
    will be returned if it's not empty.
    
    **Example 1**::
    
        >>> def some_func(name=None):
        ...     name = empty_if(name, 'John Doe')
        ...     return name
        >>> some_func("")
        John Doe
        >>> some_func("Dave")
        Dave
    
    **Example 2**::
    
        >>> empty_if(None, 'is empty', 'is not empty')
        is empty
        >>> empty_if(12345, 'is empty', 'is not empty')
        is not empty
    
    
    :param Any v:     The value to test for emptiness
    :param is_empty:  The value to return if ``v`` is empty (defaults to ``None``)
    :param not_empty: The value to return if ``v`` is not empty (defaults to the original value ``v``)
    :param kwargs:    Any additional kwargs to pass to :func:`.empty`
    
    :key zero: if ``zero=True``, then v is empty if it's int ``0`` or str ``'0'``
    :key itr:  if ``itr=True``,  then v is empty if it's ``[]``, ``{}``, or is an iterable and has 0 length
    
    :return V orig_var:  The original value ``v`` is returned if ``not_empty`` isn't specified.
    :return K is_empty:  The value specified as ``is_empty`` is returned if ``v`` is empty
    :return T not_empty: The value specified as ``not_empty`` is returned if ``v`` is not empty
                         (and not_empty was specified)
    """
    not_empty = v if not_empty == USE_ORIG_VAR else not_empty
    return is_empty if empty(v, **kwargs) else not_empty


def is_true(v) -> bool:
    """
    Check if a given bool/str/int value is some form of ``True``:
    
        * **bool**: ``True`` 
        * **str**: ``'true'``, ``'yes'``, ``'y'``, ``'1'``
        * **int**: ``1``
    
    (note: strings are automatically .lower()'d)

    Usage:

    >>> is_true('true')
    True
    >>> is_true('no')
    False

    :param Any v:          The value to check for truthfulness
    :return bool is_true:  ``True`` if the value appears to be truthy, otherwise ``False``.
    """
    v = v.lower() if type(v) is str else v
    return v in [True, 'true', 'yes', 'y', '1', 1]


def is_false(v, chk_none: bool = True) -> bool:
    """
    **Warning:** Unless you specifically need to verify a value is Falsey, it's usually safer to 
    check for truth :py:func:`.is_true` and invert the result, i.e. ``if not is_true(v)``
    
    Check if a given bool/str/int value is some form of ``False``:

        * **bool**: ``False``
        * **str**: ``'false'``, ``'no'``, ``'n'``, ``'0'``
        * **int**: ``0``

    If ``chk_none`` is True (default), will also consider the below values to be Falsey::

        boolean: None // string: 'null', 'none', ''
    
    (note: strings are automatically .lower()'d)

    Usage:

    >>> is_false(0)
    True
    >>> is_false('yes')
    False

    :param Any v:          The value to check for falseyness
    :param bool chk_none:  If ``True``, treat ``None``/``'none'``/``'null'`` as Falsey (default ``True``)
    :return bool is_False: ``True`` if the value appears to be falsey, otherwise ``False``.
    """
    v = v.lower() if type(v) is str else v
    chk = [False, 'false', 'no', 'n', '0', 0]
    chk += [None, 'none', 'null', ''] if chk_none else []
    return v in chk


def parse_keyval(line: str, valsplit: str = ':', csvsplit=',') -> List[Tuple[str, str]]:
    """
    Parses a csv with key:value pairs such as::

        John Alex:Doe,Jane Sarah:Doe
    

    Into a list with tuple pairs (can be easily converted to a dict)::

        [
            ('John Alex', 'Doe'), 
            ('Jane Sarah', 'Doe')
        ]


    By default, uses a colons ``:`` to split the key/value, and commas ``,`` to terminate the end of 
    each keyval pair. This can be overridden by changing valsplit/csvsplit.

    :param str line: A string of key:value pairs separated by commas e.g. ``John Alex:Doe,Jane Sarah:Doe``
    :param str valsplit: A character (or several) used to split the key from the value (default: colon ``:``)
    :param str csvsplit: A character (or several) used to terminate each keyval pair (default: comma ``,``)
    :return List[Tuple[str,str]] parsed_data:  A list of (key, value) tuples that can easily be casted to a dict()
    """
    cs, vs = csvsplit, valsplit
    line = [tuple(a.split(vs)) for a in line.split(cs)] if line != '' else []
    return [(a.strip(), b.strip()) for a, b in line]


def parse_csv(line: str, csvsplit: str = ',') -> List[str]:
    """
    Quick n' dirty parsing of a simple comma separated line, with automatic whitespace stripping
    of both the ``line`` itself, and the values within the commas.

    Example:
    
        >>> parse_csv('  hello ,  world, test')
        ['hello', 'world', 'test']
        >>> parse_csv(' world  ;   test   ; example', csvsplit=';')
        ['world', 'test', 'example']


    :param str line: A string of columns separated by commas e.g. ``hello,world,foo``
    :param str csvsplit: A character (or several) used to terminate each value in the list. Default: comma ``,``
    """
    return [x.strip() for x in line.strip().split(csvsplit)]


def env_csv(env_key: str, env_default = None, csvsplit=',') -> List[str]:
    """
    Quick n' dirty parsing of simple CSV formatted environment variables, with fallback
    to user specified ``env_default`` (defaults to None)

    Example:

        >>> os.setenv('EXAMPLE', '  hello ,  world, test')
        >>> env_csv('EXAMPLE', [])
        ['hello', 'world', 'test']
        >>> env_csv('NONEXISTANT', [])
        []
    
    :param str env_key:     Environment var to attempt to load
    :param any env_default: Fallback value if the env var is empty / not set (Default: None)
    :param str csvsplit: A character (or several) used to terminate each value in the list. Default: comma ``,``
    :return List[str] parsed_data: A list of str values parsed from the env var
    """
    d = env(env_key)
    return env_default if empty(d) else parse_csv(d, csvsplit=csvsplit)


def env_keyval(env_key: str, env_default=None, valsplit=':', csvsplit=',') -> List[Tuple[str, str]]:
    """
    Parses an environment variable containing ``key:val,key:val`` into a list of tuples [(key,val), (key,val)]
    
    See :py:meth:`parse_keyval`

    :param str env_key:     Environment var to attempt to load
    :param any env_default: Fallback value if the env var is empty / not set (Default: None)
    :param str valsplit: A character (or several) used to split the key from the value (default: colon ``:``)
    :param str csvsplit: A character (or several) used to terminate each keyval pair (default: comma ``,``)
    """
    d = env(env_key)
    return env_default if empty(d) else parse_keyval(d, valsplit=valsplit, csvsplit=csvsplit)


def env_cast(env_key: str, cast: callable, env_default=None):
    """
    Obtains an environment variable ``env_key``, if it's empty or not set, ``env_default`` will be returned.
    Otherwise, it will be converted into a type of your choice using the callable ``cast`` parameter

    Example:

        >>> os.environ['HELLO'] = '1.234'
        >>> env_cast('HELLO', Decimal, Decimal('0'))
        Decimal('1.234')


    :param callable cast:   A function to cast the user's env data such as ``int`` ``str`` or ``Decimal`` etc.
    :param str env_key:     Environment var to attempt to load
    :param any env_default: Fallback value if the env var is empty / not set (Default: None)
    """
    return env_default if empty(env(env_key)) else cast(env(env_key))


def env_bool(env_key: str, env_default=None) -> Union[bool, None]:
    """
    Obtains an environment variable ``env_key``, if it's empty or not set, ``env_default`` will be returned.
    Otherwise, it will be converted into a boolean using :py:func:`.is_true`

    Example:

        >>> os.environ['HELLO_WORLD'] = '1'
        >>> env_bool('HELLO_WORLD')
        True
        >>> env_bool('HELLO_NOEXIST')
        None
        >>> env_bool('HELLO_NOEXIST', 'error')
        'error'
    
    :param str env_key:     Environment var to attempt to load
    :param any env_default: Fallback value if the env var is empty / not set (Default: None)
    """
    return env_cast(env_key=env_key, cast=is_true, env_default=env_default)


def env_int(env_key: str, env_default=None) -> int:
    """Alias for :py:func:`.env_cast` with ``int`` casting"""
    return env_cast(env_key=env_key, cast=int, env_default=env_default)


def env_decimal(env_key: str, env_default=None) -> Decimal:
    """Alias for :py:func:`.env_cast` with ``Decimal`` casting"""
    return env_cast(env_key=env_key, cast=Decimal, env_default=env_default)


def dec_round(amount: Decimal, dp: int = 2, rounding=None) -> Decimal:
    """
    Round a Decimal to x decimal places using ``quantize`` (``dp`` must be >= 1 and the default dp is 2)

    If you don't specify a rounding option, it will use whatever rounding has been set in :py:func:`decimal.getcontext`
    (most python versions have this default to ``ROUND_HALF_EVEN``)

    Basic Usage:

        >>> from decimal import Decimal, getcontext, ROUND_FLOOR
        >>> x = Decimal('1.9998')
        >>> dec_round(x, 3)
        Decimal('2.000')

    Custom Rounding as an argument:

        >>> dec_round(x, 3, rounding=ROUND_FLOOR)
        Decimal('1.999')

    Override context rounding to set the default:

        >>> getcontext().rounding = ROUND_FLOOR
        >>> dec_round(x, 3)
        Decimal('1.999')


    :param Decimal   amount: The amount (as a Decimal) to round
    :param int           dp: Number of decimal places to round ``amount`` to. (Default: 2)
    :param str     rounding: A :py:mod:`decimal` rounding option, e.g. ``ROUND_HALF_EVEN`` or ``ROUND_FLOOR``
    :return Decimal rounded: The rounded Decimal amount
    """
    dp = int(dp)
    if dp <= 0:
        raise ArithmeticError('dec_round expects dp >= 1')
    rounding = getcontext().rounding if not rounding else rounding
    dp_str = '.' + str('0' * (dp - 1)) + '1'
    return Decimal(amount).quantize(Decimal(dp_str), rounding=rounding)


def chunked(iterable, n):
    """ Split iterable into ``n`` iterables of similar size

    Examples::
        >>> l = [1, 2, 3, 4]
        >>> list(chunked(l, 4))
        [[1], [2], [3], [4]]

        >>> l = [1, 2, 3]
        >>> list(chunked(l, 4))
        [[1], [2], [3], []]

        >>> l = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
        >>> list(chunked(l, 4))
        [[1, 2, 3], [4, 5, 6], [7, 8, 9], [10]]

    Taken from: https://stackoverflow.com/a/24484181/2648583

    """
    chunksize = int(math.ceil(len(iterable) / n))
    return (iterable[i * chunksize:i * chunksize + chunksize] for i in range(n))


def inject_items(items: list, dest_list: list, position: int) -> List[str]:
    """
    Inject a list ``items`` after a certain element in ``dest_list``.
    
    **NOTE:** This does NOT alter ``dest_list`` - it returns a **NEW list** with ``items`` injected after the
    given ``position`` in ``dest_list``.
    
    **Example Usage**::

        >>> x = ['a', 'b', 'e', 'f', 'g']
        >>> y = ['c', 'd']
        >>> # Inject the list 'y' into list 'x' after element 1 (b)
        >>> inject_items(y, x, 1)
        ['a', 'b', 'c', 'd', 'e', 'f', 'g']
    
    :param list items: A list of items to inject into ``dest_list``
    :param list dest_list: The list to inject ``items`` into
    :param int position: Inject ``items`` after this element (0 = 1st item) in ``dest_list``
    :return List[str] injected: :py:attr:`.dest_list` with the passed ``items`` list injected at ``position``
    """
    before = list(dest_list[0:position+1])
    after = list(dest_list[position+1:])
    return before + items + after


def byteify(data: Union[str, bytes], encoding='utf-8') -> bytes:
    """
    Convert a piece of data into bytes if it isn't already.
        
        >>> byteify("hello world")
        b"hello world"
    
    """
    return bytes(data, encoding) if type(data) is not bytes else data


def stringify(data: Union[str, bytes], encoding='utf-8') -> str:
    """
    Convert a piece of data into a string (from bytes) if it isn't already.

        >>> stringify(b"hello world")
        "hello world"

    """
    return data.decode(encoding) if type(data) is bytes else data


class ErrHelpParser(argparse.ArgumentParser):
    """
    ErrHelpParser - Use this instead of :py:class:`argparse.ArgumentParser` to automatically get full
    help output as well as the error message when arguments are invalid, instead of just an error message.

    >>> parser = ErrHelpParser(description='My command line app')
    >>> parser.add_argument('nums', metavar='N', type=int, nargs='+')

    """
    def error(self, message):
        sys.stderr.write('error: %s\n' % message)
        self.print_help()
        sys.exit(2)


def human_name(class_name: Union[str, bytes, callable, Type[object]]) -> str:
    """
    This function converts a class/function name into a Title Case name. It also directly accepts classes/functions.
    
    Input names can be either snake case ``my_function``, or InitialCaps ``MyClass`` - though mixtures of the two
    may work, such as ``some_functionName`` - however ``some_FunctionName`` will not (causes double spaces).
    
    **Examples**
    
    Using a plain string or bytes::
    
        >>> human_name(b'_some_functionName')
        'Some Function Name'
        >>> human_name('SomeClassName')
        'Some Class Name'
    
    Using a reference to a function::
    
        >>> def some_func():
        ...     pass
        >>> human_name(some_func)
        'Some Func'
    
    Using a reference to a class, or an instance of a class::
    
        >>> class MyExampleClass:
        ...     pass
        >>> my_instance = MyExampleClass()
        >>> human_name(MyExampleClass)
        'My Example Class'
        >>> human_name(my_instance)
        'My Example Class'
    
    :param class_name: The name of a class/function specified either in InitialCaps or snake_case.
                       You may also pass a function reference, class reference, or class instance. (see examples)
    :return str human_name: The humanised Title Case name of ``class_name``
    """
    # First we figure out what type ``class_name`` actually **is**.
    # Bytes simply get decoded back into a string, while strings are untouched
    if type(class_name) in [str, bytes]:
        class_name = stringify(class_name)
    # References to classes (not instances) and functions means we need .__name__
    elif type(class_name) is type or str(type(class_name)) == "<class 'function'>":
        class_name = class_name.__name__
    # If it's not a class/function reference, but is an instance of object, then it's a class instance.
    elif isinstance(class_name, object):
        class_name = class_name.__class__.__name__

    # Then we convert it into a normal string.
    class_name = str(class_name)
    
    # Strip any underlines at the start or end of the class name.
    name = class_name.strip('_').strip('-')
    # We can't alter an object as we iterate it, so we copy `name` into a new list which we'll modify instead
    new_name = list(name)
    
    # Capitalise the first letter of the name, if it isn't already.
    if name[0].islower():
        new_name[0] = name[0].upper()

    # When we inject spaces where there weren't any before, we need to track how this changes the length,
    # so that we can correctly reference positions in `new_name`
    offset = 0
    # Iterate over each character in the original name (ignoring the first letter because it's probably capital)
    for i, c in enumerate(name[1:]):
        pos = (i + 1) + offset
        # If the current character is uppercase, then inject a space before this character and increment `offset`
        if c.isupper():
            new_name = inject_items([' '], new_name, pos - 1)
            offset += 1
            continue
        # If the character is an underline or dash, replace it with a space, and uppercase the character in-front of it.
        if c in ['_', '-']:
            new_name[pos] = ' '
            if str(name[i + 2]).isalpha():
                new_name[pos + 1] = new_name[pos + 1].upper()
    
    return ''.join(new_name).strip()


IS_XARGS = re.compile('^\*([a-zA-Z0-9_])+$')
"""Pre-compiled regex for matching catch-all positional argument parameter names like ``*args``"""
IS_XKWARGS = re.compile('^\*\*([a-zA-Z0-9_])+$')
"""Pre-compiled regex for matching catch-all keyword argument parameter names like ``**args``"""
T_PARAM = inspect.Parameter
"""Type alias for :class:`inspect.Parameter`"""
T_PARAM_LIST = Union[Dict[str, T_PARAM], List[T_PARAM], Iterable[T_PARAM]]
"""
Type alias for dict's containing strings mapped to :class:`inspect.Parameter`'s, lists of just
:class:`inspect.Parameter`'s, and any iterable of :class:`inspect.Parameter`
"""

# noinspection PyProtectedMember
INS_EMPTY = inspect._empty
"""
Type alias for :class:`inspect.empty`
"""


def _filter_params(params: T_PARAM_LIST, ignore_xargs=False, ignore_xkwargs=False, **kwargs) -> Dict[str, T_PARAM]:
    """
    Filter an iterable containing :class:`inspect.Parameter`'s, returning a :class:`.DictObject` containing
    parameter names mapped to their :class:`inspect.Parameter` object.
    
    **Examples**
    
    Function ``some_func`` is used as an example.
    
        >>> import inspect
        >>> def some_func(x, y, z=123, *args, **kwargs):
        ...     pass
        >>> params = inspect.signature(some_func).parameters
    
    With just parameters, no filtering is done. Only scanning the parameters and returning them as a dict::
    
        >>> _filter_params(params)
        {'x': <Parameter "x">, 'y': <Parameter "y">, 'z': <Parameter "z=123">,
         '*args': <Parameter "*args">, '**kwargs': <Parameter "**kwargs">}
    
    With the arguments ``ignore_xargs=True`` and ``ignore_xkwargs=True``, this strips away any catch-all parameters
    e.g. ``*args`` / ``**kwargs``. Example::
    
        >>> _filter_params(params, ignore_xargs=True, ignore_xkwargs=True)
        {'x': <Parameter "x">, 'y': <Parameter "y">, 'z': <Parameter "z=123">}
    
    With the arguments ``ignore_defaults=True`` and ``ignore_positional=True``, this strips away all normal positional
    and keyword parameters - leaving only catch-all parameters for positional/keyword arguments. Example::
    
        >>> _filter_params(params, ignore_defaults=True, ignore_positional=True)
        {'*args': <Parameter "*args">, '**kwargs': <Parameter "**kwargs">}
    
    
    :param params: An iterable of :class:`inspect.Parameter`'s, e.g. from ``inspect.signature(func).parameters``
    :param bool ignore_xargs: Filter out any catch-all positional arguments (e.g. ``*args``)
    :param bool ignore_xkwargs: Filter out any catch-all keyword arguments (e.g. ``**kwargs``)
    
    :key bool ignore_defaults: Filter out any parameter which has a default value (e.g. args usable as kwargs)
    :key bool ignore_positional: Filter out any parameter which doesn't have a default value (mandatory args)
    
    :return DictObject filtered: A dictionary of filtered params, mapping param names to Parameter objects.
    """
    ignore_defaults = kwargs.pop('ignore_defaults', False)
    ignore_positional = kwargs.pop('ignore_positional', False)
    

    
    _params = params
    if isinstance(params, (dict, OrderedDict)) or hasattr(params, 'values'):
        _params = params.values()
    
    _is_xargs = lambda param: IS_XARGS.search(str(param)) is not None
    _is_xkwargs = lambda param: IS_XKWARGS.search(str(param)) is not None
    _is_x_arg = lambda param: _is_xargs(param) or _is_xkwargs(param)
    _def_empty = lambda param: empty(param.default) or param.default is INS_EMPTY
    
    filtered = DictObject()
    for p in _params:   # type: inspect.Parameter
        if ignore_xargs and _is_xargs(p): continue
        if ignore_xkwargs and _is_xkwargs(p): continue
        # x-args (*args / **kwargs) cannot count as defaults / positionals and shouldn't be counted in this IGNORE.
        if ignore_defaults and not _def_empty(p) and not _is_x_arg(p): continue
        if ignore_positional and _def_empty(p) and not _is_x_arg(p): continue
        param_name = str(p) if '=' not in str(p) else p.name
        filtered[param_name] = p
    
    return filtered


T_PARAM_DICT = Union[
    Dict[str, T_PARAM],
    DictObject,
    Dict[type, Dict[str, T_PARAM]]
]
"""
Type alias for dict's mapping parameter names to :class:`inspect.Parameter`'s, :class:`.DictObject`'s,
and dict's mapping classes to dict's mapping parameter names to :class:`inspect.Parameter`'s.
"""


def get_function_params(obj: Union[type, callable], check_parents=False, **kwargs) -> T_PARAM_DICT:
    """
    Extracts a function/method's signature (or class constructor signature if a class is passed), and returns
    it as a dictionary.
    
    Primarily used by :func:`.construct_dict` - but may be useful for other purposes.
    
    If you've passed a class, you can set ``check_parents`` to ``True`` to obtain the signatures of the passed
    class's constructor AND all of it's parent classes, returned as a dictionary mapping classes to dictionaries
    of parameters.
    
    If you've set ``check_parents`` to ``True``, but you want the parameters to be a flat dictionary (just like when
    passing a function or class without check_parents), you can also pass ``merge=True``, which merges each class's
    constructor parameters into a dictionary mapping names to :class:`inspect.Parameter` objects.
    
    If any parameters conflict, children's constructor parameters always take precedence over their parent's version,
    much in the same way that Python's inheritance works.
    
    **Basic (with functions)**::

        
        >>> def some_func(x, y, z=123, *args, **kwargs):
        ...    pass
    
    Get all normal parameters (positional and kwargs - excluding catch-all ``*args`` / ``**kwargs`` parameter types)::
    
        >>> params = get_function_params(some_func)
        >>> params
        {'x': <Parameter "x">, 'y': <Parameter "y">, 'z': <Parameter "z=123">}
    
    Get raw parameter name and value (as written in signature) / access default values::
        
        >>> str(params.z.name)     # You can also access it via params['z']
        'z=123'
        >>> params.z.default  # You can also access it via params['z']
        123

    Get only **required** parameters::
        
        >>> get_function_params(some_func, ignore_defaults=True)
        {'x': <Parameter "x">, 'y': <Parameter "y">}
    
    Get only parameters with defaults::
    
        >>> get_function_params(some_func, ignore_positional=True)
        {'z': <Parameter "z=123">}
    
    
    **Example Usage (with classes and sub-classes)**::
    
        >>> class BaseClass:
        ...     def __init__(self, a, b, c=1234, **kwargs):
        ...         pass
        
        >>> class Example(BaseClass):
        ...     def __init__(self, d, e='hello', f=None, a='overridden', **kwargs):
        ...         super().__init__(a=a, d=d, e=e, f=f, **kwargs)
    
    If we pass the class ``Example`` on it's own, we get a dictionary of just it's own parameters::
    
        >>> get_function_params(Example)
        {'d': <Parameter "d">, 'e': <Parameter "e='hello'">, 'f': <Parameter "f=None">}
    
    However, if we set ``check_parents=True``, we now get a dictionary containing ``Example``'s constructor parameters,
    AND ``BaseClass``'s (it's parent class) constructor parameters, organised by class::
        
        >>> get_function_params(Example, True)
        {
            <class '__main__.Example'>: {
                'd': <Parameter "d">, 'e': <Parameter "e='hello'">, 'f': <Parameter "f=None">,
                'a': <Parameter "a='overridden'">
            },
            <class '__main__.BaseClass'>: {'a': <Parameter "a">, 'b': <Parameter "b">, 'c': <Parameter "c=1234">}
        }
    
    We can also add the optional kwarg ``merge=True``, which merges the parameters of the originally passed class,
    and it's parents.
    
    This is done in reverse order, so that children's conflicting constructor parameters take priority over their
    parents, as can be seen below with ``a`` which is shown as ``a='overridden'`` - the overridden parameter
    of the class ``Example`` with a default value, instead of the parent's ``a`` which makes ``a`` mandatory::
     
        >>> get_function_params(Example, True, merge=True)
        {
            'a': <Parameter "a='overridden'">, 'b': <Parameter "b">, 'c': <Parameter "c=1234">,
            'd': <Parameter "d">, 'e': <Parameter "e='hello'">, 'f': <Parameter "f=None">
        }
    
    
    :param type|callable obj: A class (not an instance) or callable (function / lambda) to extract and filter the
                              parameter's from. If a class is passed, the parameters of the constructor will be
                              returned (``__init__``), excluding the initial ``self`` parameter.
    
    :param bool check_parents: (Default: ``False``) If ``obj`` is a class and this is True, will recursively grab
        the constructor parameters for all parent classes, and return the parameters as a dictionary of
        ``{<class X>: {'a': <Parameter 'a'>}, <class Y>: {'b': <Parameter 'b'>}``, unless ``merge`` is also set
        to ``True``.
    
    :key bool ignore_xargs: (Default: ``True``) Filter out any catch-all positional arguments (e.g. ``*args``)
    :key bool ignore_xkwargs: (Default: ``True``) Filter out any catch-all keyword arguments (e.g. ``**kwargs``)
    
    :key bool ignore_defaults: (Default: ``False``) Filter out any parameter which has a default
                               value (e.g. args usable as kwargs)
    
    :key bool ignore_positional: (Default: ``False``) Filter out any parameter which doesn't have a default
                                 value (mandatory args)
    
    :key bool merge: (Default: ``False``) If this is True, when ``check_parents`` is enabled, all parameters will
                     be flatted into a singular dictionary, e.g. ``{'a': <Parameter 'a'>, 'b': <Parameter "b">}``
    
    :return:
    """
    merge = kwargs.pop('merge', False)
    filter_opts = dict(**kwargs)
    filter_opts['ignore_xargs'] = filter_opts.get('ignore_xargs', True)
    filter_opts['ignore_xkwargs'] = filter_opts.get('ignore_xkwargs', True)
    
    _cls_keys = inspect.signature(obj).parameters
    cls_keys = _filter_params(inspect.signature(obj).parameters, **filter_opts)
    if check_parents and hasattr(obj, '__base__') and inspect.isclass(obj):
        ret = OrderedDictObject({obj: cls_keys})
        last_parent = obj.__base__
        while last_parent not in [None, type, object]:
            try:
                ret[last_parent] = _filter_params(
                    inspect.signature(last_parent).parameters, **filter_opts
                )
                if not hasattr(last_parent, '__base__'):
                    last_parent = None
                    continue
                last_parent = last_parent.__base__
            except (Exception, BaseException) as e:
                log.warning("Finishing check_parents loop due to exception: %s - %s", type(e), str(e))
                last_parent = None
                continue
        
        if merge:
            merged = OrderedDictObject()
            for cls in reversed(ret):
                for k, p in ret[cls].items():
                    merged[k] = p
            return merged
            
        return ret
    
    return OrderedDictObject(cls_keys)


def construct_dict(cls: Union[Type[T], C], kwargs: dict, args: Iterable = None, check_parents=True) -> Union[T, Any]:
    """
    Removes keys from the passed dict ``data`` which don't exist on ``cls`` (thus would get rejected as kwargs)
    using :func:`.get_function_params`. Then create and return an instance of ``cls``, passing the filtered
    ``kwargs`` dictionary as keyword args.

    Ensures that any keys in your dictionary which don't exist on ``cls`` are automatically filtered out, instead
    of causing an error due to unexpected keyword arguments.

    **Example - User class which only takes specific arguments**
    
    First let's define a class which only takes three arguments in it's constructor - username, first_name, last_name.

        >>> class User:
        ...    def __init__(self, username, first_name=None, last_name=None):
        ...        self.username = username
        ...        self.first_name, self.last_name = first_name, last_name
        ...
    
    Now we'll create a dictionary which has those three arguments, but also the excess ``address`` and ``phone``.
    
        >>> data = dict(username='johndoe123', first_name='John', last_name='Doe',
        ...             address='123 Example St', phone='+1-123-000-1234')
    
    If we tried to directly pass data as keyword args, we'd get an error::
     
        >>> john = User(**data)
        TypeError: __init__() got an unexpected keyword argument 'address'
    
    But by using :func:`.construct_dict`, we're able to construct a ``User``, as this helper function detects that
    the excess ``address`` and ``phone`` are not valid parameters for ``User``'s constructor.
     
        >>> from privex.helpers import construct_dict
        >>> john = construct_dict(User, data)
        >>> print(john.username, john.first_name, john.last_name)
        johndoe123 John Doe
    
    **Example - A function/method which only takes specific arguments**
    
    Not only can :func:`.construct_dict` be used for classes, but it can also be used for any function/method.
    
    Here's an example using a simple "factory function" which creates user objects::
    
        >>> def create_user(username, first_name=None, last_name=None):
        ...     return User(username, first_name, last_name)
        >>>
        >>> data = dict(
        ...     username='johndoe123', first_name='John', last_name='Doe',
        ...     address='123 Example St', phone='+1-123-000-1234'
        ... )
        >>> # We can't just pass data as kwargs due to the extra keys.
        >>> create_user(**data)
        TypeError: create_user() got an unexpected keyword argument 'address'
        >>> # But we can call the function using construct_dict, which filters out the excess dict keys :)
        >>> john = construct_dict(create_user, data)
        >>> print(john.username, john.first_name, john.last_name)
        johndoe123 John Doe
    
    
    :param Type[T]|callable cls: A class (not an instance) or callable (function / lambda) to extract and filter the
                                 parameter's from, then call using filtered ``kwargs`` and ``args``.
    
    :param dict kwargs: A dictionary containing keyword arguments to filter and use to call / construct ``cls``.
    
    :param list|set args: A list of positional arguments (NOT FILTERED!) to pass when calling/constructing ``cls``.
    
    :param bool check_parents: (Default: ``True``) If ``obj`` is a class and this is True, will recursively grab
                               the constructor parameters for all parent classes of ``cls`` and merge them into the
                               returned dict.
    :return Any func_result:   If ``cls`` was a function/method, the return result will be the returned data/object
                               from the function passed.
    
    :return T cls_instance:    If ``cls`` was a class, then the return result will be an instance of the class.
    
    """
    args = empty_if(args, [])
    if hasattr(cls, '__attrs_attrs__'):
        # If the passed object has the attribute __attrs_attrs__, then this means that it's an ``attr.s`` class, so
        # we should just extract the attributes from __attrs_attrs__.
        cls_keys = [atr.name for atr in cls.__attrs_attrs__]
    else:
        # Otherwise, extract the function / class's expected parameter names using our helper get_function_params().
        cls_keys = get_function_params(cls, check_parents=check_parents, merge=True)
        cls_keys = cls_keys.keys()

    clean_data = {x: y for x, y in kwargs.items() if x in cls_keys}
    return cls(*args, **clean_data)






"""
This module folder is just used for structuring ``helpers`` under the ``privex`` root package.

**Copyright**::

        +===================================================+
        |                 © 2019 Privex Inc.                |
        |               https://www.privex.io               |
        +===================================================+
        |                                                   |
        |        Originally Developed by Privex Inc.        |
        |                                                   |
        |        Core Developer(s):                         |
        |                                                   |
        |          (+)  Chris (@someguy123) [Privex]        |
        |          (+)  Kale (@kryogenic) [Privex]          |
        |                                                   |
        +===================================================+

    Copyright 2019     Privex Inc.   ( https://www.privex.io )

    Permission is hereby granted, free of charge, to any person obtaining a copy of 
    this software and associated documentation files (the "Software"), to deal in 
    the Software without restriction, including without limitation the rights to use, 
    copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the 
    Software, and to permit persons to whom the Software is furnished to do so, 
    subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all 
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, 
    INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A 
    PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT 
    HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION 
    OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE 
    SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

"""

__path__ = __import__('pkgutil').extend_path(__path__, __name__)




import base64
import binascii
import logging
from io import BufferedWriter, TextIOWrapper
from typing import Union, Tuple, Optional

import cryptography.exceptions
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization, hashes
from cryptography.hazmat.primitives.asymmetric import ec, rsa, padding
from cryptography.hazmat.primitives.asymmetric.ed25519 import Ed25519PrivateKey, Ed25519PublicKey
from cryptography.hazmat.primitives.asymmetric.rsa import RSAPrivateKeyWithSerialization, RSAPublicKeyWithSerialization, \
    RSAPublicKey, RSAPrivateKey
from cryptography.hazmat.primitives.hashes import HashAlgorithm
from cryptography.hazmat.primitives.serialization import PublicFormat, PrivateFormat, Encoding, load_ssh_public_key, \
    load_pem_public_key, load_pem_private_key, load_der_private_key, load_der_public_key
from cryptography.hazmat.primitives.serialization.pkcs12 import load_key_and_certificates

from privex.helpers.common import byteify, stringify, empty
from privex.helpers.crypto.base import is_base64, auto_b64decode
from privex.helpers.exceptions import InvalidFormat, EncryptionError

log = logging.getLogger(__name__)

Text = Union[str, bytes]


class Format:
    """
    Key formatting / encoding helper for the :doc:`Cryptography <cryptography:index>` module
    
    Used by :class:`.KeyManager` to simplify serialization of public/private keys.
    """
    public_formats = {
        'subject':           PublicFormat.SubjectPublicKeyInfo,
        'pkcs1':             PublicFormat.PKCS1,
        'ssh':               PublicFormat.OpenSSH,
        'raw':               PublicFormat.Raw,
        'compressedpoint':   PublicFormat.CompressedPoint,
        'uncompressedpoint': PublicFormat.UncompressedPoint
    }
    private_formats = {
        'openssl': PrivateFormat.TraditionalOpenSSL,
        'pkcs1':   PrivateFormat.TraditionalOpenSSL,
        'pkcs8':   PrivateFormat.PKCS8,
        'raw':     PrivateFormat.Raw
    }
    encodings = {
        'pem':  Encoding.PEM,
        'der':  Encoding.DER,
        'ssh':  Encoding.OpenSSH,
        'raw':  Encoding.Raw,
        'x962': Encoding.X962
    }
    private_format: PrivateFormat
    public_format: PublicFormat
    public_encoding: Encoding
    private_encoding: Encoding

    def __init__(self, private_format, private_encoding, public_format='ssh', public_encoding='ssh'):
        self.private_format = self.get_format('private_format', private_format)
        self.public_format = self.get_format('public_format', public_format)
        self.private_encoding = self.get_format('encoding', private_encoding)
        self.public_encoding = self.get_format('encoding', public_encoding)

    @classmethod
    def get_format(cls, fmtype, key) -> Union[Encoding, PublicFormat, PrivateFormat]:
        if fmtype == 'encoding':
            f = cls.encodings
            if key not in f:
                raise InvalidFormat(f'The encoding type "{key}" is not a valid type. Choices: {f.keys()}')
            return f[key]
    
        if fmtype == 'public_format':
            f = cls.public_formats
            if key not in f:
                raise InvalidFormat(f'The public format type "{key}" is not a valid type. Choices: {f.keys()}')
            return f[key]
        if fmtype == 'private_format':
            f = cls.private_formats
            if key not in f:
                raise InvalidFormat(f'The private format type "{key}" is not a valid type. Choices: {f.keys()}')
            return f[key]


class KeyManager:
    """
    Asymmetric key handling class - Generate, save, and load asymmetric keys, with signatures + encryption made easy.
    
    A wrapper around :py:mod:`cryptography.hazmat.primitives.asymmetric` to make generation, saving, loading,
    AND usage of asymmetric keys easy.
    
    **Basic Usage**
    
    Using :py:meth:`.output_keypair` - you can generate a key pair, and output it at the same time::
    
        >>> priv, pub = KeyManager.output_keypair('id_rsa', 'id_rsa.pub')
        >>> pub
        b'ssh-rsa AAAAB3NzaC1yc2EAAAADAQABAAABAQDbzCL+Dn8B9jS404mETt8fb6+TJek1afFthSBi2qZ0iL8dbv/Go0ig...'
    
    If you don't want to output the key pair to a file, you can also just generate one and have the private/public
    keys purely returned as bytes::
    
    
        >>> priv, pub = KeyManager.generate_keypair(alg='ed25519')
        >>> priv
        b'-----BEGIN PRIVATE KEY-----
        MC4CAQAwBQYDK2VwB.....T2YxW/Xkz3PkMHrrYBvI0LbUPky
        -----END PRIVATE KEY-----'
        >>> pub
        b'ssh-ed25519 AAAAC3NzaC1lZDI1NTE5AAAAICJ9OK6v2UGfCgWxzGdPlCQIps+lffLTWwuMLPdqfco6'
    
    
    **Loading a key for signature/encryption operations**
    
    You can load a private or public key directly from the string/bytes returned by :py:meth:`.generate_keypair`
    like so::
    
        >>> km = KeyManager(priv)
    
    Alternatively, you can load a key straight from disk using :py:meth:`.load_keyfile` - which will automatically
    detect the asymmetric algorithm, encoding/output format, and key type (public/private) and
    return a :class:`.KeyManager` instance::
    
        >>> km = KeyManager.load_keyfile('id_rsa.pub')
        
    **Automatic public key generation**
    
    If you load a **private key**, then the constructor will automatically generate the matching **public key** for
    you, so that you can use all signature/encryption methods available for your key algorithm.
    
    If you load a **public key**, then you will only be able to use methods which are available to public keys, such
    as :py:meth:`.verify` and :py:meth:`.encrypt` - you will NOT be able to use :py:meth:`.sign` or :py:meth:`.decrypt`
    
    **Manually accessing the public/private key class instances**
    
    If you need to access the raw cryptography PublicKey/PrivateKey instances, you can access them via the two
    attributes :py:attr:`.public_key` and :py:attr:`.private_key` after creating a KeyManager instance::
    
        >>> km.public_key
        <cryptography.hazmat.backends.openssl.rsa._RSAPublicKey object at 0x7f953848c438>
        >>> km.private_key
        <cryptography.hazmat.backends.openssl.rsa._RSAPrivateKey object at 0x7f95381c1ef0>
    
    **Signing, verification, and en/decryption**
    
    Once you have a :class:`.KeyManager` instance, you can now use the signing, verification, and en/decryption
    methods using the loaded key.
    
    Most key algorithms support signing and verification, which can be done using :py:meth:`.sign`
    and :py:meth:`.verify` respectively.
    
    Sensible values for things like padding/hash algorithms are set by default, so you can simply call :py:meth:`.sign`
    with just a message, and :py:meth:`.verify` with just the signature + message::
    
        >>> sig = km.sign('hello world')
        >>> km.verify(signature=sig, message='hello world')  # Raises InvalidSignature if it was invalid
        True
    
    With RSA keys, you can also :py:meth:`.encrypt` using the public key, and :py:meth:`.decrypt` using
    the private key::
    
        >>> msg = km.encrypt('my secret message')
        >>> km.decrypt(msg)
        b'my secret message'
    
    """
    backend = default_backend()
    
    default_gen = dict(
        rsa=dict(key_size=2048, public_exponent=65537, backend=backend),
        ecdsa=dict(curve=ec.SECP384R1(), backend=backend),
        ed25519=dict(),
    )
    """A map of key algorithms to their generator's default kwargs"""

    Format = Format
    """A class alias, allowing for access to the Format class via a class attribute"""
    default_formats = {
        'rsa': dict(private_format='pkcs8', private_encoding='pem'),
        'ecdsa': dict(private_format='pkcs8', private_encoding='pem'),
        'ed25519': dict(private_format='pkcs8', private_encoding='pem'),
    }
    """Default :class:`.Format` formatting options for serialising each key algorithm"""
    
    generators = {
        'rsa': (rsa.generate_private_key, dict(encryption_algorithm=serialization.NoEncryption())),
        'ecdsa': (ec.generate_private_key, dict(encryption_algorithm=serialization.NoEncryption())),
        'ed25519': (Ed25519PrivateKey.generate, dict(encryption_algorithm=serialization.NoEncryption())),
    }
    """
    Maps each key algorithm to a tuple containing the algorithm's generation function, and any extra kwargs
    needed for generating a private key
    """

    # noinspection PyProtectedMember
    curves = ec._CURVE_TYPES
    """An alias for Cryptography's map of string curve names (e.g. ``secp521r1``) to their respective classes"""
    
    private_key_types = Union[
        RSAPrivateKeyWithSerialization, ec.EllipticCurvePrivateKeyWithSerialization,
        Ed25519PrivateKey
    ]
    """
    The cryptography library doesn't have a standard parent type for private keys, so we need a Union to hold
    the various private key types for return types, type/instance comparison etc.
    """
    public_key_types = Union[
        RSAPublicKeyWithSerialization, ec.EllipticCurvePublicKeyWithSerialization,
        Ed25519PublicKey
    ]
    """Same as :py:attr:`.private_key_types` but for public key types."""
    
    combined_key_types = Union[private_key_types, public_key_types]
    """A Union which just combines :py:attr:`.private_key_types` and :py:attr:`.public_key_types` into one type."""
    
    raw_pub_types = public_key_types.__args__
    """This extracts the actual class types from the Union[] for isinstance() checks"""
    raw_priv_types = private_key_types.__args__
    """This extracts the actual class types from the Union[] for isinstance() checks"""

    public_key: Optional[public_key_types]
    """
    When the class is initialised, this instance attribute holds a cryptography public key class instance, or None
    if we don't have a public key available.
    """
    
    private_key: Optional[private_key_types]
    """
    When the class is initialised, this instance attribute holds a cryptography private key class instance, or None
    if we don't have a private key available.
    """
    
    type_name_map = {
        RSAPublicKey: 'rsa', RSAPrivateKey: 'rsa',
        RSAPublicKeyWithSerialization: 'rsa', RSAPrivateKeyWithSerialization: 'rsa',
        Ed25519PrivateKey: 'ed25519', Ed25519PublicKey: 'ed25519',
        ec.EllipticCurvePublicKey: 'ecdsa', ec.EllipticCurvePrivateKey: 'ecdsa',
        ec.EllipticCurvePrivateKeyWithSerialization: 'ecdsa', ec.EllipticCurvePublicKeyWithSerialization: 'ecdsa',
    }
    """Maps public/private key types to their associated algorithm name for type/instance identification"""
    
    def __init__(self, key: Union[Text, private_key_types, public_key_types], password: Text = None):
        """
        Create an instance of :class:`.KeyManager` using the public/private key data ``key``
        
        If you want to load the key from a file instead of passing it's data / key class instance, then
        you should use :py:meth:`.load_keyfile` instead to create the class instance.
        
        You do NOT need to initialize this class if you're simply using the class methods / static methods such as
        :py:meth:`.generate_keypair` or :py:meth:`.load_key` - only to use the normal instance methods which
        require a loaded public/private key, such as :py:meth:`.sign`
        
        :param key: The public/private key data, as either a string, bytes, or one of the various private key class
                    instances or public key class instances
                    (see :py:attr:`.public_key_types` and :py:attr:`.private_key_types`)
         
        :param str|bytes password: If your key data is encrypted, pass the password in this argument to decrypt it.
        """
        self.public_key, self.private_key = None, None
        k = key
        # If the key is a string, or bytes, then de-serialise the key into an object
        if type(key) in [str, bytes]:
            k, ktype = self.load_key(key, password=password)
        else:   # Otherwise, figure out whether it's a private key, or a public key.
            if isinstance(key, self.raw_priv_types):
                ktype = 'private'
            elif isinstance(key, self.raw_pub_types):
                ktype = 'public'
            else:
                raise InvalidFormat('Could not identify key type...')
        
        if ktype == 'public':
            self.public_key = k
        else:
            self.private_key = k
            self.public_key = k.public_key()
    
    def sign(self, message: Text, pad=None, hashing: HashAlgorithm = hashes.SHA256()) -> bytes:
        """
        Generate a signature for a given message using the loaded :py:attr:`.private_key`. The signature is
        Base64 encoded to allow for easy storage and transmission of the signature, and can later be verified
        by :py:meth:`.verify` using :py:attr:`.public_key`
        

        
            >>> km = KeyManager.load_keyfile('id_rsa')
            >>> sig = km.sign('hello world')        # Sign 'hello world' using the id_rsa private key
            >>> try:
            ...     km.verify(sig, 'hello world')   # Verify it using the public key (automatically generated)
            ...     print('Signature is valid')
            >>> except cryptography.exceptions.InvalidSignature:
            ...     print('Signature IS NOT VALID!')
        
        Alternatively, you can manually run :py:func:`base64.urlsafe_b64decode` to decode the signature back into
        raw bytes, then you can verify it using the ``verify`` method of a ``cryptography`` public key instance,
        such as :class:`.Ed25519PublicKey` or :class:`.RSAPublicKey`
        
        :param str|bytes message: The message to verify, e.g. ``hello world``
        :param pad: (RSA only) An instance of a cryptography padding class, e.g. :class:`.padding.PSS`
        :param HashAlgorithm hashing: (ECDSA/RSA) Use this hashing method for padding/signatures
        :raises cryptography.exceptions.InvalidSignature: When the signature does not match the message
        :return bytes sig: A base64 urlsafe encoded signature
        """
        message = byteify(message)
        if isinstance(self.private_key, Ed25519PrivateKey):
            return base64.urlsafe_b64encode(self.private_key.sign(data=message))
        
        if isinstance(self.private_key, ec.EllipticCurvePrivateKey):
            return base64.urlsafe_b64encode(self.private_key.sign(message, ec.ECDSA(hashing)))
        
        # Fallback: RSA with padding
        if not pad:
            pad = padding.PSS(mgf=padding.MGF1(hashing), salt_length=padding.PSS.MAX_LENGTH)
        return base64.urlsafe_b64encode(self.private_key.sign(data=message, padding=pad, algorithm=hashing))
    
    def verify(self, signature: Text, message: Text, pad=None, hashing: HashAlgorithm = hashes.SHA256()) -> bool:
        """
        Verify a signature against a given message using an asymmetric public key.
        
            >>> km = KeyManager.load_keyfile('id_rsa')
            >>> sig = km.sign('hello world')    # Sign 'hello world' using the id_rsa private key
            >>> try:
            ...     km.verify(sig, 'hello world')   # Verify it using the public key (automatically generated)
            ...     print('Signature is valid')
            >>> except cryptography.exceptions.InvalidSignature:
            ...     print('Signature IS NOT VALID!')
        
        :param str|bytes signature: The binary, or base64 urlsafe encoded signature to check ``message`` against
        :param str|bytes message: The message to verify, e.g. ``hello world``
        :param pad: (RSA only) An instance of a cryptography padding class, e.g. :class:`.padding.PSS`
        :param HashAlgorithm hashing: (ECDSA/RSA) Use this hashing method for padding/signatures
        :raises cryptography.exceptions.InvalidSignature: When the signature does not match the message
        :return bool is_valid: ``True`` if signature is valid, otherwise raises InvalidSignature.
        """
        signature, message = auto_b64decode(signature), byteify(message)
        
        if isinstance(self.public_key, Ed25519PublicKey):
            return self.public_key.verify(signature=signature, data=message)
        
        if isinstance(self.public_key, ec.EllipticCurvePublicKey):
            return self.public_key.verify(signature=signature, data=message, signature_algorithm=ec.ECDSA(hashing))
        
        # Fallback: RSA with padding
        if not pad:
            pad = padding.PSS(mgf=padding.MGF1(hashing), salt_length=padding.PSS.MAX_LENGTH)
        self.public_key.verify(signature=signature, data=message, padding=pad, algorithm=hashing)
        return True
    
    def encrypt(self, message: Text, pad=None, hashing: HashAlgorithm = hashes.SHA256()) -> bytes:
        """
        Encrypt a message using the loaded :py:attr:`.public_key` - returns the ciphertext as base64 encoded
        bytes.
        
        NOTE: **Only works for RSA public keys**

        :param str|bytes message: The message to encrypt, e.g. ``hello world``
        :param pad: (RSA only) An instance of a cryptography padding class, e.g. :class:`.padding.PSS`
        :param HashAlgorithm hashing: (ECDSA/RSA) Use this hashing method for padding/signatures
        :return bytes ciphertext: The encrypted version of ``message`` - encoded with Base64 for easy storage/transport
        """
        message = byteify(message)
        
        if not isinstance(self.public_key, (RSAPublicKey, RSAPublicKeyWithSerialization)):
            raise InvalidFormat('KeyManager.encrypt is only supported for RSA keys.')
        if not pad:
            pad = padding.OAEP(mgf=padding.MGF1(hashing), algorithm=hashing, label=None)
        return base64.urlsafe_b64encode(self.public_key.encrypt(plaintext=message, padding=pad))

    def decrypt(self, message: Text, pad=None, hashing: HashAlgorithm = hashes.SHA256()) -> bytes:
        """
        Decrypt a message using the loaded :py:attr:`.private_key` - returns the decrypted message as bytes.
        
        NOTE: **Only works for RSA private keys**
        
        :param str|bytes message: The ciphertext to decrypt, as base64 or raw bytes
        :param pad: (RSA only) An instance of a cryptography padding class, e.g. :class:`.padding.PSS`
        :param HashAlgorithm hashing: (ECDSA/RSA) Use this hashing method for padding/signatures
        :return bytes decoded: The decrypted version of ``message`` as bytes
        """
        message = auto_b64decode(message)
    
        if not isinstance(self.private_key, (RSAPrivateKey, RSAPrivateKeyWithSerialization)):
            raise InvalidFormat('KeyManager.decrypt is only supported for RSA keys.')
        if not pad:
            pad = padding.OAEP(mgf=padding.MGF1(hashing), algorithm=hashing, label=None)
        return self.private_key.decrypt(ciphertext=message, padding=pad)
    
    @classmethod
    def identify_algorithm(cls, key: combined_key_types) -> str:
        """
        Identifies a cryptography public/private key instance, such as :class:`.RSAPrivateKey`
        and returns the algorithm name that can be used with other KeyManager methods,
        e.g. ``'rsa'`` or ``'ed25519'``
        
        
        Example::
            
            >>> priv, pub = KeyManager.generate_keypair_raw(alg='ecdsa', curve=ec.SECP521R1)
            >>> KeyManager.identify_algorithm(priv)
            'ecdsa'
            >>> priv, pub = KeyManager.generate_keypair_raw()
            >>> KeyManager.identify_algorithm(priv)
            'rsa'
            >>> KeyManager.identify_algorithm(pub)
            'rsa'


        :param combined_key_types key: A cryptography public/private key instance
        :return str algorithm: The name of the algorithm used by this key
        """
        for ktype, kname in cls.type_name_map.items():
            if isinstance(key, ktype):
                return kname
        raise InvalidFormat('Could not identify key type...')

    def export_public(self, **kwargs) -> bytes:
        """
        Serialize the cryptography public key instance loaded into :class:`.KeyManager` into storable bytes.
        
        This method works whether you've instantiated KeyManager with the public key directly, or the private key,
        as the public key is automatically interpolated from the private key by :py:meth:`.__init__`

        Example::
        
            >>> km = KeyManager.load_keyfile('id_ed25519.pub')
            >>> km.export_public()
            b'ssh-ed25519 AAAAC3NzaC1lZDI1NTE5AAAAIA6vtKgeNSBERSY1xmr47Ve3uyRALxPR+qOeFeUHrUaf'
            
        :keyword dict format: Override some or all of the default format/encoding for the keys.
                              Dict Keys: private_format,public_format,private_encoding,public_encoding
        :keyword Format format: If passed a :class:`.Format` instance, then this instance will be used for serialization
                                instead of merging defaults from :py:attr:`.default_formats`
        :return bytes key: The serialized key.
        """
        if empty(self.public_key):
            raise EncryptionError('Cannot export public key as self.public_key is missing!')
        # alg = self.identify_algorithm(self.public_key)
        return self.export_key(self.public_key, **kwargs)

    def export_private(self, **kwargs) -> bytes:
        """
        Serialize the cryptography private key instance loaded into :class:`.KeyManager` into storable bytes.
        
        This method requires that you've instantiated KeyManager with the private key. It will raise a
        :class:`.EncryptionError` exception if the :py:attr:`.private_key` instance attribute is empty.
        
        Example::

            >>> km = KeyManager.load_keyfile('id_ed25519')
            >>> print(km.export_private().decode())
            -----BEGIN PRIVATE KEY-----
            MC4CAQAwBQYDK2VwBCIEIOeLS2XOcQz11VUnzh6KIZaNtT10YfzHv779zjm95XSy
            -----END PRIVATE KEY-----


        :keyword dict format: Override some or all of the default format/encoding for the keys.
                              Dict Keys: private_format,public_format,private_encoding,public_encoding
        :keyword Format format: If passed a :class:`.Format` instance, then this instance will be used for serialization
                                instead of merging defaults from :py:attr:`.default_formats`
        :return bytes key: The serialized key.
        """
        if empty(self.private_key):
            raise EncryptionError('Cannot export private key as self.private_key is missing!')
        # alg = self.identify_algorithm(self.private_key)
        return self.export_key(self.private_key, **kwargs)
    
    @staticmethod
    def _load_der_key(data: bytes, password: Optional[bytes] = None):
        """Attempt to de-serialise a DER formatted private/public key. Raises :class:`.InvalidFormat` if it fails."""
        try:
            key = load_der_private_key(data, password=password, backend=default_backend())
            return key, 'private'
        except (ValueError, TypeError, cryptography.exceptions.UnsupportedAlgorithm):
            pass
        try:
            key = load_der_public_key(data, backend=default_backend())
            return key, 'public'
        except (ValueError, TypeError, cryptography.exceptions.UnsupportedAlgorithm):
            raise InvalidFormat('Not a DER key, or is corrupted...')
    
    @staticmethod
    def _load_pkcs12_key(data: bytes, password: Optional[bytes] = None):
        """Attempt to de-serialise PKCS12 key data. Raises :class:`.InvalidFormat` if it fails."""
        try:
            key, cert, ad_certs = load_key_and_certificates(data, password=password, backend=default_backend())
            return key, 'private', cert, ad_certs
        except (ValueError, TypeError, cryptography.exceptions.UnsupportedAlgorithm):
            raise InvalidFormat('Not a PKCS12 key, or is corrupted...')
    
    @classmethod
    def load_key(cls, data: Text, password: bytes = None) -> Tuple[combined_key_types, str]:
        """
        Load a private/public key from a string or bytes ``data`` containing the key in some format, such as PEM
        or OpenSSH. Use :py:meth:`.load_keyfile` to load a key from a file + auto-instantiate KeyManager with it.
        
        **Example:**
        
            >>> key = 'ssh-ed25519 AAAAC3NzaC1lZDI1NTE5AAAAIG3v6guHpI/j7AIl3F/EWMSReX8fH8MOSq1bC3ZuEjjC'
            >>> KeyManager.load_key(key)
            (<cryptography.hazmat.backends.openssl.ed25519._Ed25519PublicKey object at 0x7fa118289ba8>, 'public')
        
        
        :param str|bytes data: The public/private key data, as a string or bytes
        :param str|bytes password: If your key data is encrypted, pass the password in this argument to decrypt it.
        :raises InvalidFormat: When the key could not be identified, is not supported, or is corrupted
        :return tuple key_data: A tuple containing an instance of the key, and a string ``public`` or ``private``
                                Example: ``(<_Ed25519PublicKey object>, 'public')``
        """
        data = stringify(data)
        password = None if not password else byteify(password)
        try:
            if data[0:4] == 'ssh-' or data[0:6] == 'ecdsa-':
                return load_ssh_public_key(byteify(data), backend=default_backend()), 'public'
            if '-----BEGIN PUBLIC' in data:
                return load_pem_public_key(byteify(data), backend=default_backend()), 'public'
            if '-----BEGIN PRIVATE' in data:
                return load_pem_private_key(byteify(data), password=password, backend=default_backend()), 'private'
        except binascii.Error as e:
            raise InvalidFormat(f'Error while decoding key - possibly corrupted base64 encoding. '
                                f'Original exception: {type(e)} {str(e)}')
        except ValueError as e:
            raise InvalidFormat(f'Error while decoding key - key may be corrupted. '
                                f'Original exception: {type(e)} {str(e)}')
        ##
        # If we couldn't identify the key as PEM or OpenSSH, then fallback to attempting to load it as DER / PKCS12
        ##
        data = byteify(data)
        
        try:   # Try loading the key as a DER key
            key, ktype = cls._load_der_key(data=data, password=password)
            return key, ktype
        except InvalidFormat:
            pass

        try:   # Try loading the key as PKCS12
            key, ktype, _, _ = cls._load_pkcs12_key(data=data, password=password)
            return key, ktype
        except InvalidFormat:
            pass
        
        # Otherwise we have no idea what to do with this key. Time to give up.
        raise InvalidFormat('Unknown key format...')

    @classmethod
    def load_keyfile(cls, filename: Text, password: Text = None):
        """
        Returns an instance of :class:`.KeyManager` using a public/private key loaded from disk, instead of
        from string/bytes key data.
        
        Example::
        
            >>> km = KeyManager.load_keyfile('id_rsa')
            >>> d = km.encrypt('hello world')
            >>> km.decrypt(d)
            b'hello world'
        
        
        :param str|bytes filename: The file location where the key is stored
        :param str|bytes password: If the key is encrypted, specify the password to decrypt it
        :raises InvalidFormat: When the key could not be identified, is not supported, or is corrupted
        :raises FileNotFoundError: The given ``filename`` couldn't be found.
        :return KeyManager cls: An instance of :class:`.KeyManager` (or child class) initialised with the key
        """
        filename = stringify(filename)
        with open(filename, 'r') as fp:
            data = fp.read()
        return cls(key=data, password=password)
    
    @classmethod
    def generate_keypair_raw(cls, alg='rsa', **kwargs) -> Tuple[private_key_types, public_key_types]:
        """
        Generate a key pair, returning private + public key instances from the cryptography module.
        
        Example::
        
            >>> priv, pub = KeyManager.generate_keypair_raw(alg='rsa', key_size=1024)
            >>> priv.private_bytes(encoding=Encoding.PEM, format=PrivateFormat.PKCS8)
            b'-----BEGIN PRIVATE KEY-----\\nMIICdQIBADANBgkqhkiG9w0BAQEFAASCAl8wggJbAgEAAoGBAMjkl
             ...Pw6eZGFwBEYY\\n-----END PRIVATE KEY-----\\n'
             
            >>> priv, pub = KeyManager.generate_keypair_raw(alg='ecdsa', curve=ec.SECP521R1)
            >>> pub.public_bytes(encoding=Encoding.OpenSSH, format=PublicFormat.OpenSSH)
            b'ecdsa-sha2-nistp521 AAAAE2VjZHNhLXNoYTItbmlzdHA1...dJCxguBQnb1hL6aDH4fHCjpy6A=='
        
        
        
        :param str alg: The algorithm to generate a key for, e.g. ``'rsa'``
        :param kwargs: All kwargs are forwarded to the matching generator in :py:attr:`.generators`
        :return tuple keys: A tuple containing a private key instance, and public key instance
        """
        gen_args = {**cls.default_gen[alg], **kwargs}
        gen = cls.generators[alg][0]
    
        priv = gen(**gen_args)
        pub = priv.public_key()
        return priv, pub
    
    @classmethod
    def generate_keypair(cls, alg='rsa', **kwargs) -> Tuple[bytes, bytes]:
        """
        Generate a key pair, returning private + public key as serialized bytes based on :py:attr:`.default_formats`
        and the kwarg ``format`` if it's present.
        
        By default, private keys are generally returned in PKCS8 format with PEM encoding, while
        public keys are OpenSSH format and OpenSSH encoding.
        
        Example::
        
            >>> priv, pub = KeyManager.generate_keypair(alg='rsa', key_size=2048)
            >>> priv
            b'-----BEGIN PRIVATE KEY-----\\nMIICdQIBADANBgkqhkiG9w0BAQEFAASCAl8wggJbAgEAAoGBAMjkl
             ...Pw6eZGFwBEYY\\n-----END PRIVATE KEY-----\\n'
             
            >>> priv, pub = KeyManager.generate_keypair(alg='ecdsa', curve=ec.SECP521R1)
            >>> pub
            b'ecdsa-sha2-nistp521 AAAAE2VjZHNhLXNoYTItbmlzdHA1...dJCxguBQnb1hL6aDH4fHCjpy6A=='
        
        
        To override the formatting/encoding::
        
            >>> priv, pub = KeyManager.generate_keypair(
            ...     alg='ecdsa', format=dict(private_format='openssl', private_encoding='der')
            ... )
            >>> priv
            b'0\\x81\\xa4\\x02\\x01\\x01\\x040u\\x1e\\x8cI\\xcd\\xfa\\xc8\\x97\\x83\\xf8\\xed\\x1f\\xe5\\xbd...'
        
        
        :param str alg: The algorithm to generate a key for, e.g. ``'rsa'``
        :param kwargs: All kwargs are forwarded to the matching generator in :py:attr:`.generators`
        :keyword int key_size: (for `rsa` and similar algorithms) Number of bits for the RSA key. Minimum of 512 bits.
        :keyword dict format: Override some or all of the default format/encoding for the keys.
                      Dict Keys: private_format,public_format,private_encoding,public_encoding
        :return:
        """
        # noinspection PyTypeChecker
        fmt = kwargs.pop('format') if 'format' in kwargs else {}
        
        priv, pub = cls.generate_keypair_raw(alg=alg, **kwargs)

        s_priv = cls.export_key(priv, format=fmt, alg=alg)
        s_pub = cls.export_key(pub, format=fmt, alg=alg)

        return s_priv, s_pub

    @classmethod
    def export_key(cls, key: combined_key_types, **kwargs) -> bytes:
        """
        Export/serialize a given public/private key object as bytes.
        
        Uses the default formatting arguments for the detected algorithm from :py:meth:`.identify_algorithm`, but
        you can also force it to treat it as a certain algorithm by passing ``alg``
        
        Uses default formatting options by looking up the algorithm in :py:attr:`.default_formats`
        
        Uses the private key serialization arguments for the detected algorithm out of :py:attr:`.generators`
        
        Example::
            
            >>> priv, pub = KeyManager.generate_keypair_raw('ed25519')
            >>> key = KeyManager.export_key(pub)
            >>> print(key.decode())
            -----BEGIN PRIVATE KEY-----
            MC4CAQAwBQYDK2VwBCIEIOeLS2XOcQz11VUnzh6KIZaNtT10YfzHv779zjm95XSy
            -----END PRIVATE KEY-----
        
        :param str alg: An algorithm name as a string, e.g. ``rsa`` or ``ed25519``
        :param combined_key_types key: An instance of a public/private key type listed in :py:attr:`.combined_key_types`
        :keyword dict format: Override some or all of the default format/encoding for the keys.
                              Dict Keys: private_format,public_format,private_encoding,public_encoding
        :keyword Format format: If passed a :class:`.Format` instance, then this instance will be used for serialization
                                instead of merging defaults from :py:attr:`.default_formats`
        :keyword str alg: Use this algorithm name e.g. ``'rsa'`` instead of detecting using :meth:`.identify_algorithm`
        :return bytes key: The serialized key.
        """
        alg = kwargs.pop('alg', cls.identify_algorithm(key))
        fmt_args = cls.default_formats[alg]
        priv_args = cls.generators[alg][1]
        _fmt = kwargs.get('format', {})
        fmt: Format = _fmt if isinstance(_fmt, Format) else Format(**{**fmt_args, **_fmt})

        if isinstance(key, cls.raw_priv_types):
            return key.private_bytes(encoding=fmt.private_encoding, format=fmt.private_format, **priv_args)
        elif isinstance(key, cls.raw_pub_types):
            return key.public_bytes(encoding=fmt.public_encoding, format=fmt.public_format)
        else:
            raise InvalidFormat(f'export_key expected a cryptography private/public key instance. got: {type(key)}')

    @classmethod
    def output_keypair(cls, priv: Union[str, BufferedWriter], pub: Union[str, TextIOWrapper],
                       *args, **kwargs) -> Tuple[bytes, bytes]:
        """
        Similar to :py:meth:`.generate_keypair` - except this also writes the private key and public key to the
        file locations and/or byte streams specified in the first two arguments (``priv`` and ``pub``)
        
        **Example**
        
        Generate a 4096-bit RSA key pair, and output the private key to the file ``id_rsa`` ,
        and the public key to ``id_rsa.pub``.
        
        The generated keypair is also returned as a tuple pair (bytes) containing the private and public key::
        
            >>> priv, pub = KeyManager.output_keypair('id_rsa', 'id_rsa.pub', alg='rsa', key_size=4096)
        
        
        :param str|BufferedWriter priv: File location to save private key, or writable byte stream
        :param str|BufferedWriter pub: File location to save public key, or writable byte stream
        :param args: All additional args are forwarded to :py:meth:`.generate_keypair`
        :param kwargs: All kwargs are forwarded to :py:meth:`.generate_keypair`
        
        """
        privkey, pubkey = cls.generate_keypair(*args, **kwargs)
        
        if isinstance(priv, BufferedWriter):
            priv.write(privkey)
        else:
            with open(str(priv), 'wb') as fp:
                fp.write(privkey)

        if isinstance(pub, BufferedWriter):
            pub.write(pubkey)
        else:
            with open(str(pub), 'wb') as fp:
                fp.write(pubkey)
        
        return privkey, pubkey



"""
Functions/classes which depend on the package ``attrs`` (`Attrs.org`_)

.. _Attrs.org: https://www.attrs.org/en/stable/

**Copyright**::

        +===================================================+
        |                 © 2019 Privex Inc.                |
        |               https://www.privex.io               |
        +===================================================+
        |                                                   |
        |        Originally Developed by Privex Inc.        |
        |        License: X11 / MIT                         |
        |                                                   |
        |        Core Developer(s):                         |
        |                                                   |
        |          (+)  Chris (@someguy123) [Privex]        |
        |          (+)  Kale (@kryogenic) [Privex]          |
        |                                                   |
        +===================================================+

    Copyright 2019     Privex Inc.   ( https://www.privex.io )


"""
import attr

__all__ = [
    'AttribDictable'
]


@attr.s
class AttribDictable:
    """
    A small mixin class for :py:func:`attr.s` classes, allowing them to behave like dictionaries, and
    support casting into dictionaries using ``dict(x)``.
    
    Usage::
        
        >>> import attr
        >>> from privex.helpers import AttribDictable
        >>>
        >>> @attr.s
        >>> class Example(AttribDictable):
        ...     hello = attr.ib(type=str)
        ...     testing = attr.ib(type=bool, default=True)
        ...
        >>> x = Example(hello='world')
        >>> x['hello']
        'world'
        >>> x['hello'] = 'lorem ipsum'
        >>> dict(x)
        {'hello': 'lorem ipsum', 'testing': True}
        
    
    """
    def get(self, key, default=None):
        try:
            return self[key]
        except KeyError:
            return default
    
    def __iter__(self):
        """Handle casting via ``dict(myclass)``"""
        for k, v in attr.asdict(self).items():
            yield k, v
    
    def __getitem__(self, key):
        """
        When the instance is accessed like a dict, try returning the matching attribute.
        If the attribute doesn't exist, raise :class:`KeyError`
        """
        if hasattr(self, key):
            return getattr(self, key)
        raise KeyError(key)
    
    def __setitem__(self, key, value):
        return setattr(self, key, value)



"""
Helpers for setup.py, e.g. requirements.txt parsing, version bumping, custom setup.py commands


Inside of :py:mod:`privex.helpers.setuppy.common` there's a variety of functions related to generating
requirements.txt files, parsing requirements.txt files which recursively import other requirements.txt files,
and handing automatic generation of ``extras_require`` from a folder containing requirements txt files.

Inside of :py:mod:`privex.helpers.setuppy.bump` - most notably is :py:func:`.bump_version` - a function which
detects a package's version, increments the appropriate part of the version number, and then updates the python
file containing the version number (e.g. an ``__init__.py``)


Inside of :py:mod:`privex.helpers.setuppy.commands` there are command classes which can be loaded into setup.py to
assist with building python packages, generating requirements.txt files from extras, as well as general management
such as a :class:`.BumpCommand` which allows you to bump your package version with a simple ``./setup.py bump --minor``

More detailed usage documentation is available within each individual module's documentation.

"""

try:
    from privex.helpers.setuppy.common import *
except ImportError:
    log.debug('%s failed to import "setuppy.common", not loading ...', __name__)
try:
    from privex.helpers.setuppy.bump import *
except ImportError:
    log.debug('%s failed to import "setuppy.bump", not loading ...', __name__)
try:
    from privex.helpers.setuppy.commands import *
except ImportError:
    log.debug('%s failed to import "setuppy.commands", not loading ...', __name__)



"""
Helper functions/classes related to caching.

This module acts as a singleton wrapper, allowing for easily setting a framework-independent global cache API.

To make the module easy to use, :py:func:`.adapter_get` initialises an instance of :class:`.MemoryCache` if no
global cache adapter instance has been setup. This means you can use the various alias functions in this module
without having to configure a cache adapter.

Setting / updating the global cache adapter instance
----------------------------------------------------

First import the ``cache`` module.

    >>> from privex.helpers import cache

You must instantiate your cache adapter of choice before passing it to :py:func:`.adapter_set` - which updates
the global cache adapter instance.

    >>> my_adapter = cache.MemoryCache()
    >>> cache.adapter_set(my_adapter)

Once you've set the adapter, you can use the module functions such as :py:func:`.get` and :py:func:`.set` - or you
can import ``cached`` to enable dictionary-like cache item access.

    >>> cache.set('hello', 'world')
    >>> cache.get('hello')
    'world'
    >>> from privex.helpers import cached
    >>> cached['hello']
    'world'
    >>> cached['otherkey'] = 'testing'


Plug-n-play usage
-----------------

As explained near the start of this module's documentation, you don't have to set the global adapter if you only
plan on using the simple :class:`.MemoryCache` adapter.

Just start using the global cache API via either :py:mod:`privex.helpers.cache` or :py:mod:`privex.helpers.cache.cached`
and MemoryCache will automatically be instantiated as the global adapter as soon as something attempts to access
the global instance.

We recommend importing ``cached`` rather than ``cache``, as it acts as a wrapper that allows dictionary-like
cache key getting/setting, and is also immediately aware when the global cache adapter is set/replaced.

    >>> from privex.helpers import cached

You can access ``cached`` like a dictionary to get and set cache keys (they will use the default expiry time of
:py:attr:`privex.helpers.settings.DEFAULT_CACHE_TIMEOUT`)

    >>> cached['testing'] = 123
    >>> cached['testing']
    123


You can also call methods such as :py:func:`.get` and :py:func:`.set` for getting/setting cache items with more
control, for example:

1. Setting a custom expiration, or disabling expiration by setting timeout to ``None``
  
    >>> cached.set('example', 'test', timeout=30)   # Drop 'example' from the cache after 30 seconds from now.
    >>> cached.set('this key', 'is forever!', timeout=None) # A timeout of ``None`` disables automatic expiration.

2. Fallback values when a key isn't found, or have it throw an exception if it's not found instead.

    >>> cached.get('example', 'NOT FOUND')          # If the key 'example' doesn't exist, return 'NOT FOUND'
    'test'
    
    >>> try:   # By setting ``fail`` to True, ``get`` raises ``CacheNotFound`` if the key doesn't exist / is expired
    ...     cached.get('nonexistent', fail=True)
    ... except CacheNotFound:
    ...     log.error('The cache key "nonexistent" does not exist!')
    >>>

3. Using :py:func:`.get_or_set` you can specify either a standard type (e.g. ``str``, ``int``, ``dict``), or even
   a custom function to call to obtain the value to set and return.
 
    >>> cached.get_or_set('hello', lambda key: 'world', timeout=60)
    >>> cached['hello']
    'world'


**Copyright**::

        +===================================================+
        |                 © 2019 Privex Inc.                |
        |               https://www.privex.io               |
        +===================================================+
        |                                                   |
        |        Originally Developed by Privex Inc.        |
        |        License: X11 / MIT                         |
        |                                                   |
        |        Core Developer(s):                         |
        |                                                   |
        |          (+)  Chris (@someguy123) [Privex]        |
        |          (+)  Kale (@kryogenic) [Privex]          |
        |                                                   |
        +===================================================+

    Copyright 2019     Privex Inc.   ( https://www.privex.io )


"""
import logging

from typing import Any, Optional, Union, Type

from privex.helpers.cache.CacheAdapter import CacheAdapter
from privex.helpers.cache.MemoryCache import MemoryCache

try:
    from privex.helpers.cache.RedisCache import RedisCache
except ImportError:
    pass

from privex.helpers.exceptions import NotConfigured, CacheNotFound
from privex.helpers.settings import DEFAULT_CACHE_TIMEOUT

log = logging.getLogger(__name__)

__STORE = {}

__STORE['adapter']: CacheAdapter


class CacheWrapper(object):
    """
    **CacheWrapper** is a small class designed to wrap an instance of :class:`.CacheAdapter` and allow
    the adapter to be switched out at any time, using the static class attribute :py:attr:`.cache_instance`.
    
    This class is used for the singleton global variable :py:attr:`.cached`
    
    For convenience, if :py:attr:`.cache_instance` isn't set-up when something makes an adapter-dependant call, then
    the adapter class in :py:attr:`.default_adapter` will be instantiated and stored in :py:class:`.cache_instance`
    
        >>> # Using the ``: CacheAdapter`` type hinting will allow most IDEs to treat the wrapper as if it were
        >>> # a normal CacheAdapter child class, thus showing appropriate completion / usage warnings
        >>> c: CacheAdapter = CacheWrapper()
        >>> c.set('hello', 'world')
        >>> c['hello']
        'world'
    
    You can replace the cache adapter singleton using the module function :py:func:`.adapter_set` (recommended)
    
        >>> from privex.helpers import cache, CacheWrapper
        >>> cache.adapter_set(cache.MemoryCache())   # Set the current adapter for both the cache module, and wrapper.

    If you only plan to use this wrapper, then you can use :py:meth:`.set_adapter` to update the current cache adapter
    instance.
    
        >>> CacheWrapper.set_adapter(cache.MemoryCache())  # Set the adapter only for the wrapper (aka ``cached``)
    
    """
    cache_instance: CacheAdapter = None
    """Holds the singleton instance of a :class:`.CacheAdapter` implementation"""
    
    default_adapter: Type[CacheAdapter] = MemoryCache
    """The default adapter class to instantiate if :py:attr:`.cache_instance` is ``None``"""
    
    @staticmethod
    def get_adapter(default: Type[CacheAdapter] = default_adapter, *args, **kwargs) -> CacheAdapter:
        """
        Attempt to get the singleton cache adapter from :py:attr:`.cache_instance` - if the instance is ``None``, then
        attempt to instantiate ``default()``
        
        If any ``*args`` or ``**kwargs`` are passed, they will be passed through to ``default(*args, **kwargs)`` so
        that any necessary configuration parameters can be passed to the class.
        """
        if not CacheWrapper.cache_instance:
            CacheWrapper.cache_instance = default(*args, **kwargs)
        return CacheWrapper.cache_instance

    @staticmethod
    def set_adapter(adapter: CacheAdapter) -> CacheAdapter:
        CacheWrapper.cache_instance = adapter
        return CacheWrapper.cache_instance

    def __getattr__(self, item):
        if hasattr(super(), item):
            return getattr(self, item)
        
        def _wrapper(*args, **kwargs):
            return getattr(CacheWrapper.get_adapter(), item)(*args, **kwargs)
        
        return _wrapper
    
    def __getitem__(self, item):
        try:
            return CacheWrapper.get_adapter().get(key=item, fail=True)
        except CacheNotFound:
            raise KeyError(f'Key "{item}" not found in cache.')
    
    def __setitem__(self, key, value):
        return CacheWrapper.get_adapter().set(key=key, value=value)


cached: CacheAdapter = CacheWrapper()
"""
This module attribute acts as a singleton, containing an instance of :class:`.CacheWrapper` which is designed
to allow painless usage of this caching module.


"""


def adapter_set(adapter: CacheAdapter):
    """
    Set the global cache adapter instance to ``adapter`` - which should be an instantiated adapter class which
    implements :class:`.CacheAdapter`
    
    **Example**::
    
        >>> from privex.helpers import cache
        >>> cache.adapter_set(cache.MemoryCache())
    
    
    :param CacheAdapter adapter: An instance of a class which implements :class:`.CacheAdapter` for global use.
    :return CacheAdapter adapter: A reference to your adapter from ``__STORE['adapter']``
    """
    __STORE['adapter'] = adapter
    cached.set_adapter(adapter)
    
    return __STORE['adapter']


def adapter_get(default: Type[CacheAdapter] = MemoryCache) -> CacheAdapter:
    """
    Get the global cache adapter instance. If there isn't one, then by default this function will initialise
    :class:`.MemoryAdapter` and set it as the global cache adapter.
    
    To set the global cache adapter instance, use :py:func:`.adapter_set`
    
    To use a different fallback class, pass a class name which implements :class:`.CacheAdapter` like so:
    
        >>> adapter_get(default=MemoryCache)
    
    :param default:
    :return:
    """
    if 'adapter' not in __STORE or __STORE['adapter'] is None:
        if not default:
            raise NotConfigured('No cache adapter has been configured for privex.helpers.cache!')
        __STORE['adapter'] = default()
    return __STORE['adapter']


def get(key: str, default: Any = None, fail: bool = False) -> Any:
    """
    Return the value of cache key ``key``. If the key wasn't found, or it was expired, then ``default`` will be
    returned.
    
    Optionally, you may choose to pass ``fail=True``, which will cause this method to raise :class:`.CacheNotFound`
    instead of returning ``default`` when a key is non-existent / expired.
    
    :param str key: The cache key (as a string) to get the value for, e.g. ``example:test``
    :param Any default: If the cache key ``key`` isn't found / is expired, return this value (Default: ``None``)
    :param bool fail: If set to ``True``, will raise :class:`.CacheNotFound` instead of returning ``default``
                      when a key is non-existent / expired.
    
    :raises CacheNotFound: Raised when ``fail=True`` and ``key`` was not found in cache / expired.
    
    :return Any value: The value of the cache key ``key``, or ``default`` if it wasn't found.
    """
    a = adapter_get()
    return a.get(key=key, default=default, fail=fail)


def set(key: str, value: Any, timeout: Optional[int] = DEFAULT_CACHE_TIMEOUT):
    """
    Set the cache key ``key`` to the value ``value``, and automatically expire the key after ``timeout`` seconds
    from now.
    
    If ``timeout`` is ``None``, then the key will never expire (unless the cache implementation loses it's
    persistence, e.g. memory caches with no disk writes).
    
    :param str key: The cache key (as a string) to set the value for, e.g. ``example:test``
    :param Any value: The value to store in the cache key ``key``
    :param int timeout: The amount of seconds to keep the data in cache. Pass ``None`` to disable expiration.
    """
    a = adapter_get()
    return a.set(key=key, value=value, timeout=timeout)


def get_or_set(key: str, value: Union[Any, callable], timeout: int = DEFAULT_CACHE_TIMEOUT) -> Any:
    """
    Attempt to return the value of ``key`` in the cache. If ``key`` doesn't exist or is expired, then it will be
    set to ``value``, and ``value`` will be returned.
    
    The ``value`` parameter can be any standard type such as ``str`` or ``dict`` - or it can be a callable
    function / method which returns the value to set and return.
    
    **Basic Usage**::
    
        >>> from privex.helpers import cache as c
        >>> c.get('testing')
        None
        >>> c.get_or_set('testing', 'hello world')
        'hello world'
        >>> c.get('testing')
        'hello world'
    
    **Set and get the value from a function if ``key`` didn't exist / was expired**::
    
        >>> def my_func(): return "hello world"
        >>> c.get_or_set('example', my_func)
        'hello world'
        >>> c.get('example')
        'hello world'
    
    :param str key: The cache key (as a string) to get/set the value for, e.g. ``example:test``
    :param Any value: The value to store in the cache key ``key``. Can be a standard type, or a callable function.
    :param int timeout: The amount of seconds to keep the data in cache. Pass ``None`` to disable expiration.
    :return Any value: The value of the cache key ``key``, or ``value`` if it wasn't found.
    """
    a = adapter_get()
    return a.get_or_set(key=key, value=value, timeout=timeout)


def remove(*key: str) -> bool:
    """
    Remove one or more keys from the cache.

    If all cache keys existed before removal, ``True`` will be returned. If some didn't exist (and thus couldn't
    remove), then ``False`` will be returned.

    :param str key: The cache key(s) to remove
    :return bool removed: ``True`` if ``key`` existed and was removed
    :return bool removed: ``False`` if ``key`` didn't exist, and no action was taken.
    """
    a = adapter_get()
    return a.remove(*key)


def update_timeout(key: str, timeout: int = DEFAULT_CACHE_TIMEOUT) -> Any:
    """
    Update the timeout for a given ``key`` to ``datetime.utcnow() + timedelta(seconds=timeout)``

    This method allows keys which are already expired, allowing expired cache keys to have their timeout
    extended **after** expiry.

    **Example**::

        >>> from privex.helpers import cache
        >>> from time import sleep
        >>> cache.set('example', 'test', timeout=60)
        >>> sleep(70)
        >>> cache.update_timeout('example', timeout=60)   # Reset the timeout for ``'example'`` to ``now + 60 seconds``
        >>> cache.get('example')
        'test'

    :param str key: The cache key to update the timeout for
    :param int timeout: Reset the timeout to this many seconds from ``datetime.utcnow()``
    :raises CacheNotFound: Raised when ``key`` was not found in cache (thus cannot extend timeout)
    :return Any value: The value of the cache key
    """
    a = adapter_get()
    return a.update_timeout(key=key, timeout=timeout)




"""
Various functions/classes which convert/parse objects from one type into another.

**Copyright**::

        +===================================================+
        |                 © 2019 Privex Inc.                |
        |               https://www.privex.io               |
        +===================================================+
        |                                                   |
        |        Originally Developed by Privex Inc.        |
        |        License: X11 / MIT                         |
        |                                                   |
        |        Core Developer(s):                         |
        |                                                   |
        |          (+)  Chris (@someguy123) [Privex]        |
        |          (+)  Kale (@kryogenic) [Privex]          |
        |                                                   |
        +===================================================+

    Copyright 2019     Privex Inc.   ( https://www.privex.io )


"""
from datetime import datetime
from decimal import Decimal
from typing import Optional, Union


from privex.helpers.common import empty, is_true, stringify
import logging

log = logging.getLogger(__name__)

MINUTE = 60
HOUR = MINUTE * 60
DAY = HOUR * 24
MONTH = DAY * 30
YEAR = DAY * 365
DECADE = YEAR * 10


def convert_datetime(d, if_empty=None, fail_empty=False, **kwargs) -> Optional[datetime]:
    """
    Convert the object ``d`` into a :class:`datetime.datetime` object.
    
    If ``d`` is a string or bytes, then it will be parsed using :func:`dateutil.parser.parse`
    
    If ``d`` is an int/float/Decimal, then it will be assumed to be a unix epoch timestamp.
    
    **Examples**::
        
        >>> convert_datetime("2019-01-01T00:00:00Z")          # ISO date/time
        datetime.datetime(2019, 1, 1, 0, 0, tzinfo=tzutc())
        
        >>> convert_datetime("01/JAN/2019 00:00:00.0000")     # Human date/time with month name
        datetime.datetime(2019, 1, 1, 0, 0, tzinfo=tzutc())
        
        >>> convert_datetime(1546300800)                      # Unix timestamp as integer
        datetime.datetime(2019, 1, 1, 0, 0, tzinfo=tzutc())
        
        >>> convert_datetime(1546300800000)                   # Unix timestamp (milliseconds) as integer
        datetime.datetime(2019, 1, 1, 0, 0, tzinfo=tzutc())
    
    :param d: Object to convert into a datetime
    :param if_empty: If ``d`` is empty / None, return this value
    :param bool fail_empty: (Def: ``False``) If this is True, then if ``d`` is empty, raises :class:`AttributeError`
    
    :key datetime.tzinfo tzinfo: (Default: :class:`dateutil.tz.tzutc`) If no timezone was detected by the parser,
                                 use this timezone. Set this to ``None`` to disable forcing timezone-aware dates.
    
    :raises AttributeError: When ``d`` is empty and ``fail_empty`` is set to True.
    :raises dateutil.parser.ParserError: When ``d`` could not be parsed into a date.
    :return datetime converted: The converted :class:`datetime.datetime` object.
    """
    from dateutil.tz import tzutc
    _tzinfo = kwargs.pop('tzinfo', tzutc())
    if isinstance(d, datetime):
        if d.tzinfo is None and _tzinfo is not None:
            d = d.replace(tzinfo=_tzinfo)
        return d

    d = stringify(d) if isinstance(d, bytes) else d
    
    if isinstance(d, (int, float)):
        return convert_unixtime_datetime(d)
    
    if isinstance(d, str):
        from dateutil.parser import parse, ParserError
        try:
            t = parse(d)
            if t.tzinfo is None and _tzinfo is not None:
                t = t.replace(tzinfo=_tzinfo)
            return t
        except (ParserError, ValueError) as e:
            log.info("Failed to parse string. Attempting to parse as unix time")
            try:
                t = convert_unixtime_datetime(d)
                return t
            except (BaseException, Exception, ParserError) as _err:
                log.warning("Failed to parse unix time. Re-raising original parser error. Unixtime error was: %s %s",
                            type(_err), str(_err))
                raise e
        except ImportError as e:
            msg = "ERROR: Could not import 'parse' from 'dateutil.parser'. Please " \
                  f"make sure 'python-dateutil' is installed. Exception: {type(e)} - {str(e)}"
            
            log.exception(msg)
            raise ImportError(msg)
    if empty(d):
        if fail_empty: raise AttributeError("Error converting datetime. Parameter 'd' was empty!")
        return if_empty
    if not isinstance(d, datetime):
        raise ValueError('Timestamp must be either a datetime object, or an ISO8601 string...')
    return d


def convert_unixtime_datetime(d: Union[str, int, float, Decimal], if_empty=None, fail_empty=False) -> datetime:
    """Convert a unix timestamp into a :class:`datetime.datetime` object"""
    from dateutil.tz import tzutc
    if empty(d):
        if fail_empty: raise AttributeError("Error converting datetime. Parameter 'd' was empty!")
        return if_empty
    if isinstance(d, datetime):
        return d
    d = int(d)
    # If the timestamp is larger than NOW + 50 years in seconds, then it's probably milliseconds.
    if d > datetime.utcnow().timestamp() + (DECADE * 5):
        t = datetime.utcfromtimestamp(d // 1000)
    else:
        t = datetime.utcfromtimestamp(d)
    
    t = t.replace(tzinfo=tzutc())
    return t


def convert_bool_int(d, if_empty=0, fail_empty=False) -> int:
    """Convert a boolean ``d`` into an integer (``0`` for ``False``, ``1`` for ``True``)"""
    if type(d) is int: return 1 if d >= 1 else 0
    if empty(d):
        if fail_empty: raise AttributeError(f"Error converting '{d}' into a boolean. Parameter 'd' was empty!")
        return if_empty
    return 1 if is_true(d) else 0


def convert_int_bool(d, if_empty=False, fail_empty=False) -> bool:
    """Convert an integer ``d`` into a boolean (``0`` for ``False``, ``1`` for ``True``)"""
    if empty(d):
        if fail_empty: raise AttributeError(f"Error converting '{d}' into a boolean. Parameter 'd' was empty!")
        return if_empty
    return is_true(d)


__all__ = [
    'convert_datetime', 'convert_unixtime_datetime', 'convert_bool_int', 'convert_int_bool',
    'MINUTE', 'HOUR', 'DAY', 'MONTH', 'YEAR', 'DECADE',
]




"""
Class Method / Function decorators

**Copyright**::

        +===================================================+
        |                 © 2019 Privex Inc.                |
        |               https://www.privex.io               |
        +===================================================+
        |                                                   |
        |        Originally Developed by Privex Inc.        |
        |        License: X11 / MIT                         |
        |                                                   |
        |        Core Developer(s):                         |
        |                                                   |
        |          (+)  Chris (@someguy123) [Privex]        |
        |          (+)  Kale (@kryogenic) [Privex]          |
        |                                                   |
        +===================================================+

    Copyright 2019     Privex Inc.   ( https://www.privex.io )


"""
import functools
import logging
from enum import Enum
from time import sleep
from typing import Any, Union

from privex.helpers.cache import cached
from privex.helpers.common import empty


DEF_RETRY_MSG = "Exception while running '%s', will retry %d more times."
DEF_FAIL_MSG = "Giving up after attempting to retry function '%s' %d times."

log = logging.getLogger(__name__)

try:
    from privex.helpers.asyncx import async_sync
except ImportError:
    log.debug('privex.helpers __init__ failed to import "asyncx", not loading async helpers')
    pass


def retry_on_err(max_retries: int = 3, delay: Union[int, float] = 3, **retry_conf):
    """
    Decorates a function or class method, wraps the function/method with a try/catch block, and will automatically
    re-run the function with the same arguments up to `max_retries` time after any exception is raised, with a
    ``delay`` second delay between re-tries.

    If it still throws an exception after ``max_retries`` retries, it will log the exception details with ``fail_msg``,
    and then re-raise it.

    Usage (retry up to 5 times, 1 second between retries, stop immediately if IOError is detected):

        >>> @retry_on_err(5, 1, fail_on=[IOError])
        ... def my_func(self, some=None, args=None):
        ...     if some == 'io': raise IOError()
        ...      raise FileExistsError()

    This will be re-ran 5 times, 1 second apart after each exception is raised, before giving up:

        >>> my_func()

    Where-as this one will immediately re-raise the caught IOError on the first attempt, as it's passed in ``fail_on``:

        >>> my_func('io')


    :param int max_retries:  Maximum total retry attempts before giving up
    :param float delay:      Amount of time in seconds to sleep before re-trying the wrapped function
    :param retry_conf:       Less frequently used arguments, pass in as keyword args:

    - (list) fail_on:  A list() of Exception types that should result in immediate failure (don't retry, raise)

    - (str) retry_msg: Override the log message used for retry attempts. First message param %s is func name,
      second message param %d is retry attempts remaining

    - (str) fail_msg:  Override the log message used after all retry attempts are exhausted. First message param %s
      is func name, and second param %d is amount of times retried.

    """
    retry_msg = retry_conf['retry_msg'] if 'retry_msg' in retry_conf else DEF_RETRY_MSG
    fail_msg = retry_conf['fail_msg'] if 'fail_msg' in retry_conf else DEF_FAIL_MSG
    fail_on = list(retry_conf['fail_on']) if 'fail_on' in retry_conf else []

    def _decorator(f):
        @functools.wraps(f)
        def wrapper(*args, **kwargs):
            retries = 0
            if 'retry_attempts' in kwargs:
                retries = int(kwargs['retry_attempts'])
                del kwargs['retry_attempts']

            try:
                return f(*args, **kwargs)
            except Exception as e:
                if type(e) in fail_on:
                    log.warning('Giving up. Re-raising exception %s (as requested by `fail_on` arg)', type(e))
                    raise e
                if retries < max_retries:
                    log.info('%s - %s', type(e), str(e))
                    log.info(retry_msg, f.__name__, max_retries - retries)
                    sleep(delay)
                    kwargs['retry_attempts'] = retries + 1
                    return wrapper(*args, **kwargs)
                log.exception(fail_msg, f.__name__, max_retries)
                raise e
        return wrapper
    return _decorator


class FormatOpt(Enum):
    """
    This enum represents various options available for :py:func:`.r_cache` 's ``format_opt`` parameter.

    To avoid bloating the PyDoc for ``r_cache`` too much, descriptions for each formatting option is available as a
    short PyDoc comment under each enum option.

    Usage:

    >>> @r_cache('mykey', format_args=[0, 'x'], format_opt=FormatOpt.POS_AUTO)

    """
    POS_AUTO = 'force_pos'
    """
    First attempt to format using ``*args`` whitelisted in ``format_args``, if that causes a KeyError/IndexError,
    then pass kwarg values in the order they're listed in ``format_args``
    (only includes kwarg names listed in ``format_args``)

    # def func(x, y)
    func('a', 'b')      # assuming 0 and 1 are in format_args, then it would use .format('a', 'b')
    func(y='b', x='a')  # assuming format_args = ``['x','y']``, then it would use .format('a', 'b')
    """

    POS_ONLY = 'pos_only'
    """Only use positional args for formatting the cache key, kwargs will be ignored completely."""

    KWARG_ONLY = 'kwarg'
    """Only use kwargs for formatting the cache key - requires named format placeholders, i.e. ``mykey:{x}``"""

    MIX = 'mix'
    """
    Use both ``*args`` and ``**kwargs`` to format the cache_key (assuming mixed placeholders e.g. ``mykey:{}:{y}``
    """


FO = FormatOpt


def r_cache(cache_key: Union[str, callable], cache_time=300, format_args: list = None,
            format_opt: FO = FO.POS_AUTO, **opts) -> Any:
    """
    This is a decorator which caches the result of the wrapped function with the global cache adapter from
    :py:mod:`privex.helpers.cache` using the key ``cache_key`` and with an expiry of ``cache_time`` seconds.

    Future calls to the wrapped function would then load the data from cache until the cache expires, upon which it
    will re-run the original code and re-cache it.

    To bypass the cache, pass kwarg ``r_cache=False`` to the wrapped function. To override the cache key on demand,
    pass ``r_cache_key='mykey'`` to the wrapped function.

    **Example usage**:

        >>> from privex.helpers import r_cache
        >>>
        >>> @r_cache('mydata', cache_time=600)
        ... def my_func(*args, **kwargs):
        ...     time.sleep(60)
        ...     return "done"

        This will run the function and take 60 seconds to return while it sleeps

        >>> my_func()
        done

        This will run instantly because "done" is now cached for 600 seconds

        >>> my_func()
        done

        This will take another 60 seconds to run because ``r_cache`` is set to `False` (disables the cache)

        >>> my_func(r_cache=False)
        done

    **Using a dynamic cache_key**:

        **Simplest and most reliable - pass ``r_cache_key`` as an additional kwarg**

        If you don't mind passing an additional kwarg to your function, then the most reliable method is to override
        the cache key by passing ``r_cache_key`` to your wrapped function.

        Don't worry, we remove both ``r_cache`` and ``r_cache_key`` from the kwargs that actually hit your function.

        >>> my_func(r_cache_key='somekey')    # Use the cache key 'somekey' when caching data for this function

        **Option 2. Pass a callable which takes the same arguments as the wrapped function**

        In the example below, ``who`` takes two arguments: ``name`` and ``title`` - we then pass the function
        ``make_key`` which takes the same arguments - ``r_cache`` will detect that the cache key is a function
        and call it with the same ``(*args, **kwargs)`` passed to the wrapped function.

        >>> from privex.helpers import r_cache
        >>>
        >>> def make_key(name, title):
        ...     return f"mycache:{name}"
        ...
        >>> @r_cache(make_key)
        ... def who(name, title):
        ...     return "Their name is {title} {name}"
        ...

        We can also obtain the same effect with a ``lambda`` callable defined directly inside of the cache_key.

        >>> @r_cache(lambda name,title: f"mycache:{name}")
        ... def who(name, title):
        ...     return "Their name is {title} {name}"

        **Option 3. Can be finnicky - using ``format_args`` to integrate with existing code**

        If you can't change how your existing function/method is called, then you can use the ``format_args`` feature.

        **NOTE:** Unless you're forcing the usage of kwargs with a function/method, it's strongly recommended that you
        keep ``force_pos`` enabled, and specify both the positional argument ID, and the kwarg name.

        Basic Example:

        >>> from privex.helpers import r_cache
        >>> import time
        >>>
        >>> @r_cache('some_cache:{}:{}', cache_time=600, format_args=[0, 1, 'x', 'y'])
        ... def some_func(x=1, y=2):
        ...     time.sleep(5)
        ...     return 'x + y = {}'.format(x + y)
        >>>

        Using positional arguments, we can see from the debug log that it's formatting the ``{}:{}`` in the key
        with ``x:y``

        >>> some_func(1, 2)
        2019-08-21 06:58:29,823 lg  DEBUG    Trying to load "some_cache:1:2" from cache
        2019-08-21 06:58:29,826 lg  DEBUG    Not found in cache, or "r_cache" set to false. Calling wrapped function.
        'x + y = 3'
        >>> some_func(2, 3)
        2019-08-21 06:58:34,831 lg  DEBUG    Trying to load "some_cache:2:3" from cache
        2019-08-21 06:58:34,832 lg  DEBUG    Not found in cache, or "r_cache" set to false. Calling wrapped function.
        'x + y = 5'

        When we passed ``(1, 2)`` and ``(2, 3)`` it had to re-run the function for each. But once we re-call it for
        the previously ran ``(1, 2)`` - it's able to retrieve the cached result just for those args.

        >>> some_func(1, 2)
        2019-08-21 06:58:41,752 lg  DEBUG    Trying to load "some_cache:1:2" from cache
        'x + y = 3'

        Be warned that the default format option ``POS_AUTO`` will make kwargs' values be specified in the same order as
        they were listed in ``format_args``

        >>> some_func(y=1, x=2)   # ``format_args`` has the kwargs in the order ``['x', 'y']`` thus ``.format(x,y)``
        2019-08-21 06:58:58,611 lg  DEBUG    Trying to load "some_cache:2:1" from cache
        2019-08-21 06:58:58,611 lg  DEBUG    Not found in cache, or "r_cache" set to false. Calling wrapped function.
        'x + y = 3'

    :param bool whitelist: (default: ``True``) If True, only use specified arg positions / kwarg keys when formatting
                           ``cache_key`` placeholders. Otherwise, trust whatever args/kwargs were passed to the func.
    :param FormatOpt format_opt: (default: :py:attr:`.FormatOpt.POS_AUTO`) "Format option" - how should args/kwargs be
                                 used when filling placeholders in the ``cache_key`` (see comments on FormatOption)
    :param list format_args: A list of positional arguments numbers (e.g. ``[0, 1, 2]``) and/or kwargs
                             ``['x', 'y', 'z']`` that should be used to format the `cache_key`
    :param str cache_key: The cache key to store the cached data into, e.g. `mydata`
    :param int cache_time: The amount of time in seconds to cache the result for (default: 300 seconds)
    :return Any res: The return result, either from the wrapped function, or from the cache.
    """
    fmt_args = [] if not format_args else format_args
    r = cached
    whitelist = opts.get('whitelist', True)

    def format_key(args, kwargs):
        pos_args = args
        kw_args = kwargs
        if whitelist:
            # Whitelisted positional arguments
            pos_args = [args[i] for i in fmt_args if type(i) is int and len(args) > i]
            # Whitelisted keyword args, as a dict
            kw_args = {i: kwargs[i] for i in fmt_args if type(i) is str and i in kwargs}

        if format_opt == FormatOpt.POS_AUTO:
            log.debug('Format: POS_AUTO - Formatting with *args, fallback on err to positional **kwargs values')
            try:
                log.debug('Attempting to format with args: %s', pos_args)
                rk = cache_key.format(*pos_args)
            except (KeyError, IndexError):
                pos_kwargs = [v for _, v in kw_args.items()]
                log.debug('Failed to format with pos args, now trying positional kwargs: %s', pos_kwargs)
                rk = cache_key.format(*pos_kwargs)
            return rk

        if format_opt == FormatOpt.KWARG_ONLY:  # Only attempt to format cache_key using kwargs
            log.debug('Format: KWARG_ONLY - Formatting using only **kwargs')
            return cache_key.format(**kw_args)

        if format_opt == FormatOpt.POS_ONLY:  # Only attempt to format cache_key using positional args
            log.debug('Format: POS_ONLY - Formatting using only *args')
            return cache_key.format(*pos_args)

        if format_opt == FormatOpt.MIX:  # Format cache_key with both positional and kwargs as-is
            log.debug('Format: MIX - Formatting using passthru *args and **kwargs')
            return cache_key.format(*pos_args, **kw_args)

    def _decorator(f):
        @functools.wraps(f)
        def wrapper(*args, **kwargs):

            # Extract r_cache and r_cache_key from the wrapped function's kwargs if they're specified,
            # then remove them from the kwargs so they don't interfere with the wrapped function.
            enable_cache, rk = kwargs.get('r_cache', True), kwargs.get('r_cache_key', cache_key)
            if 'r_cache' in kwargs: del kwargs['r_cache']
            if 'r_cache_key' in kwargs: del kwargs['r_cache_key']

            if callable(rk):
                rk = rk(*args, **kwargs)
            elif not empty(fmt_args, itr=True) or not whitelist:
                # If the cache key contains a format placeholder, e.g. {somevar} - then attempt to replace the
                # placeholders using the function's kwargs
                log.debug('Format_args not empty (or whitelist=False), formatting cache_key "%s"', cache_key)
                rk = format_key(args, kwargs)
            log.debug('Trying to load "%s" from cache', rk)
            data = r.get(rk)

            if empty(data) or not enable_cache:
                log.debug('Not found in cache, or "r_cache" set to false. Calling wrapped function.')
                data = f(*args, **kwargs)
                r.set(rk, data, timeout=cache_time)

            return data

        return wrapper

    return _decorator


def mock_decorator(*dec_args, **dec_kwargs):
    """
    This decorator is a pass-through decorator which does nothing other than be a decorator.
    
    It's designed to be used with the :class:`privex.helpers.common.Mocker` class when mocking classes/modules,
    allowing you to add fake decorators to the mock class/method which do nothing, other than act like a decorator
    without breaking your functions/methods.
    """
    def _decorator(f):
        @functools.wraps(f)
        def wrapper(*args, **kwargs):
            return f(*args, **kwargs)
        return wrapper
    return _decorator




"""
Automated Python package version bumping

Summary
-------

Included is a standalone function :py:func:`.bump_version` - which when called, loads the file
:py:attr:`.settings.VERSION_FILE` , extracts the current version, bumps the requested part of the version,
then replaces the version line inside of the file so that it contains the newly bumped version.

There's also a setup.py command class :py:class:`.BumpCommand` which allows you to use :py:func:`.bump_version`
as a ``setup.py`` command, making it simple to increment your package's version without having to manually edit
the file.

If the package :py:mod:`semver` is detected, the version bumping helper :py:func:`.bump_version` becomes available.

If the module :py:mod:`distutils` is detected, the setup.py command class :py:class:`.BumpCommand` also becomes
available.


How to version your package for best compatibility
--------------------------------------------------

To avoid having to write your own version detection / replacement functions, we recommend placing your package
version inside of a python module file, such as ``__init__.py`` - as the string variable ``VERSION``

Example, ``mypackage/__init__.py``


.. code-block:: python

    from mypackage.somemodule import x
    from mypackage.othermodule import y

    VERSION = '1.2.3'


If you cannot store your package version this way for some reason, then you can write a custom detection/replacement
function and register it as the default.

See the docs for :py:func:`.bump_version` to learn how to write a custom version detection/replacement function.


Using the BumpCommand distutils command in your setup.py
--------------------------------------------------------

For :class:`.BumpCommand` to function, you must at least set :py:attr:`privex.helpers.settings.VERSION_FILE`
to an absolute path to the file which contains your package version attribute.

If your package version isn't defined as shown in the previous section on how to version your package,
then you'll **also need to set up custom version detection/replacement functions**.
(see docs at :py:func:`.bump_version`)

Below is an example ``setup.py`` file, which configures the VERSION_FILE setting to point to ``mypackage/__init__.py``
relative to the folder setup.py is in, then calls :py:func:`setuptools.setup` with the package version and command
dictionary.

.. code-block:: python

    from os import join, dirname, abspath
    from setuptools import setup, find_packages
    from privex.helpers import settings, BumpCommand
    # If you placed your version in your package __init__ then you can import it for use in setup.py building
    from mypackage import VERSION

    # This results in an absolute path to the folder where this setup.py file is contained
    BASE_DIR = dirname(abspath(__file__))

    # The file which contains "VERSION = '1.2.3'"
    settings.VERSION_FILE = join(BASE_DIR, 'mypackage', '__init__.py')

    # Register BumpCommand as a command in your setup() function.
    setup(
        version=VERSION,
        cmdclass={
            'bump': BumpCommand
        },
    );


Basic usage of the bump command with setup.py
---------------------------------------------

Once you've configured  :py:attr:`privex.helpers.settings.VERSION_FILE` and registered the command class in the
``setup()`` function, you can now use the ``bump`` command from your ``setup.py`` and it will automatically bump
your version.

Below is an example of basic usage. If you need more help on usage, type ``./setup.py bump --help`` - or for
detailed documentation on the command, see the class documentation :class:`.BumpCommand`

.. code-block:: bash

    ./setup.py bump --patch
    # Bumping 'patch' version part
    # Updating version stored in file @ /tmp/helpers/privex/helpers/__init__.py
    # Package version has been bumped from 2.0.0 to 2.0.1 and written to the file
    # /tmp/helpers/privex/helpers/__init__.py

    ./setup.py bump --minor
    # ... version has been bumped from 2.0.0 to 2.1.0 and written to the file ...


"""
import re
import logging
import semver
from privex.helpers import settings
try:
    from privex.helpers.setuppy.commands import BumpCommand
except ImportError:
    pass


log = logging.getLogger(__name__)

_find_ver = re.compile(r'VERSION = \'?\"?([0-9a-zA-Z-._]+)\'?\"?')
"""Regex used for finding/replacing version line"""


def version_replace(data: str, old_version: str, new_version: str) -> str:
    """
    Replace the version line in ``data`` containing ``old_version`` with a version line containing ``new_version``
    
    Example::
    
        >>> data = "# Example\\nVERSION = '1.2.3' # Some comment"
        >>> version_replace(data=data, old_version='1.2.3', new_version='1.3.0')
        "# Example\\nVERSION = '1.3.0' # Some comment"
    
    
    As shown in the above example, it shouldn't affect surrounding lines, or even in-line comments.
    
    Note: ``old_version`` isn't really used by this function. It exists for compatibility with user drop-in
    replacement functions that may need to know the old version to replace it.
    
    :param str data: The string contents containing ``VERSION = 'x.y.z'``
    :param str old_version: The existing version number, e.g. ``'1.2.3'``
    :param str new_version: The new version to replace it with, e.g. ``'1.3.0'``
    :return str replaced: ``data`` with the VERSION line updated.
    """
    return _find_ver.sub(f"VERSION = '{new_version}'", data)


def get_current_ver(data: str = None):
    return str(_find_ver.search(data).group(1))


default_replace_func = version_replace
"""If no version replacement function is passed to :py:func:`.bump_version`, then this function will be used."""

default_current_ver = get_current_ver
"""If no version retrieval function is passed to :py:func:`.bump_version`, then this function will be used."""


def bump_version(part='patch', dry=False, **kwargs):
    """
    Bump semver version and replace version line inside of :py:attr:`.settings.VERSION_FILE`
    
     * Obtains the current package version using ``version_func``
     * Uses :py:mod:`semver` to increment the ``part`` portion and resets any lower portions to zero
     * Reads the file :py:attr:`.settings.VERSION_FILE` and passes it to ``replace_func`` along with the original
       version and new bumped version to obtain the modified file contents.
     * Writes the file contents containing the updated version number back to :py:attr:`.settings.VERSION_FILE`
    
    Basic usage:
    
        >>> from privex.helpers import settings, setuppy
        >>> bump_version('minor')
    
    If you want to use this function outside of privex-helpers, for your own package/project, ensure you
    adjust the settings and version functions as required.
    
    To change the file which contains your package version, as well as the function used to get the
    current version::
    
        >>> import mypackage
        >>> from privex.helpers import settings
        
        >>> settings.VERSION_FILE = '/home/john/mypackage/mypackage/__init__.py'
        
    If you use the same version line format at privex-helpers, like this::
    
        VERSION = '1.2.3'   # In-line comments are fine, as are double quotes instead of single quotes.
    
    Then you don't need to make a custom version retrieval or replacement function.
    
    Otherwise... this is how you write and register a custom version retrieval and replacement function:
    
    .. code-block:: python
       :force:
        
        import re
        from privex.helpers.setuppy import bump
        
        # Regex to find the string: version='x.y.z'
        # and extract the version x.y.z on it's own.
        my_regex = re.compile(r'version=\'?\"?([0-9a-zA-Z-._]+)\'?\"?')

        def my_version_finder(data: str):
            return str(my_regex.search(data).group(1))
        
        # Set your function `my_version_finder` as the default used to obtain the current package version
        bump.default_current_ver = my_version_finder
        
        def my_version_replacer(data: str, old_version: str, new_version: str):
            # This is an example of a version replacer if you just put your version straight into setup.py
            return data.replace(f"version='{old_version}'", f"version='{new_version}'")
            
            # Alternatively use regex substitution
            return my_regex.sub(f"version='{new_version}'", data)
        
        # Set your function `my_version_replacer` as the default used to replace the version in a file.
        bump.default_replace_func = my_version_replacer
        
        
    
    :param bool dry: If set to ``True``, will only return the modified file contents instead of overwriting
                     the file :py:attr:`.settings.VERSION_FILE`
    :param str part: The part of the version to bump: ``patch``, ``minor``, ``major``, ``build`` or ``prerelease``
    :key callable replace_func: Custom version replacement function. Should take the arguments
                                (data, old_version, new_version) and return ``data`` with the version line replaced.
    :key callable version_func: Custom version retrieval function. Takes no args, returns curr version as a string.
    :key str token: If using part ``build`` or ``prerelease``, this overrides the version token
    :return:
    """
    replace_func = kwargs.get('replace_func', default_replace_func)
    get_version = kwargs.get('version_func', default_current_ver)

    ver_path = settings.VERSION_FILE
    log.debug('Reading file %s to replace version line', ver_path)
    with open(ver_path) as fp:
        ver_file = str(fp.read())

    curr_ver = get_version(ver_file)
    new_ver = _bump_version(version=curr_ver, part=part, **kwargs)
    log.debug('Current version: %s ||| Bumped version: %s', curr_ver, new_ver)
    
    new_ver_file = replace_func(data=ver_file, old_version=curr_ver, new_version=new_ver)
    if dry:
        log.debug('Dry kwarg was True. Returning modified file instead of outputting it.')
        return new_ver_file

    log.debug('Attempting to write updated contents back to %s', ver_path)
    with open(ver_path, 'w') as fp:
        fp.write(new_ver_file)
    
    return new_ver, curr_ver


def _bump_version(version: str, part: str, **kwargs) -> str:
    """
    Bumps the semver part ``part`` in the version ``version`` and returns it as a string.
    
    Used internally by :py:func:`.bump_version`
    
        >>> _bump_version('1.2.3', 'minor')
        '1.3.0'
    
    :param version: The version to bump as a string e.g. ``'1.2.3'```
    :param part: The part of the version to bump, e.g. ``minor`` or ``major``
    :key str token: If using part ``build`` or ``prerelease``, this overrides the version token
    :raises AttributeError: If ``part`` isn't a supported version part.
    :return str new_ver: The bumped version number
    """
    bumps = dict(
        minor=semver.bump_minor,
        major=semver.bump_major,
        patch=semver.bump_patch,
        build=semver.bump_build,
        prerelease=semver.bump_prerelease,
        pre=semver.bump_prerelease,
    )
    if part not in bumps:
        raise AttributeError(f'The "part" argument must be one of the following: {",".join(bumps.keys())}')
    ver_args = dict(version=version)
    if 'token' in kwargs:
        ver_args['token'] = kwargs['token']
        print('_bump_version token:', ver_args['token'])
    new_ver = bumps[part](**ver_args)
    return new_ver





"""
Various helper functions/classes which depend on a certain package being installed.

This constructor file attempts to load each extras module individually, each wrapped with a try/catch
for :class:`ImportError` so that one unavailable package doesn't cause problems.

"""
import logging

log = logging.getLogger(__name__)

__all__ = ['HAS_ATTRS']
HAS_ATTRS = False

try:
    from privex.helpers.extras.attrs import *
    HAS_ATTRS = True
    __all__ += ['AttribDictable']
except ImportError:
    log.debug('privex.helpers.extras __init__ failed to import "attrs", not loading attrs library helpers')
    



